import os

import numpy as np

from pypetal.fromfile.run_toml import make_toml
from pypetal.utils.petalio import write_data


def write_bash(toml_fname, env1, env2):

    """Need to use virtualenv. This won't work otherwise."""

    txt = """source {}/bin/activate
python run_toml.py "pre" {}
deactivate

source {}/bin/activate
python run_toml.py "jav" {}
deactivate

source {}/bin/activate
python run_toml.py "post" {}
deactivate
""".format( env1, toml_fname, env2, toml_fname, env1, toml_fname )



    with open('run.sh', 'w+') as f:
        f.write(txt)

    return



def bash_from_python(output_dir, arg2,
                    toml_fname, env1, env2,
                    line_names=None,
                    run_drw_rej=False, drw_rej_params={},
                    run_detrend=False, detrend_params={},
                    run_pyccf=False, pyccf_params={},
                    run_pyzdcf=False, pyzdcf_params={},
                    run_pyroa=False, pyroa_params={},
                    run_javelin=False, javelin_params={},
                    run_weighting=False, weighting_params={},
                    **kwargs):

    if arg2 is None:
        raise Exception('Please provide a list of light curve filenames or the light curves themselves')


    if not isinstance(arg2[0], str):
        os.makedirs( output_dir + 'input_lcs/', exist_ok=True )
        fnames = []

        for i in range( len(arg2) ):

            if i == 0:
                name = 'continuum'
            else:
                name = 'line{}'.format(i+1)

            write_data( arg2[i], output_dir + 'input_lcs/' + name + '.dat' )
            fnames.append( output_dir + 'input_lcs/' + name + '.dat' )

        fnames = np.array(fnames)
        kwargs['file_fmt'] = 'csv'
    else:
        fnames = arg2



    run_arr = [run_drw_rej, run_detrend, run_pyccf, run_pyzdcf, run_pyroa, run_javelin, run_weighting]
    param_arr = [kwargs, drw_rej_params, detrend_params, pyccf_params, pyzdcf_params, pyroa_params, javelin_params, weighting_params]

    make_toml(output_dir, fnames, run_arr, param_arr, line_names, filename=toml_fname)
    write_bash(toml_fname, env1, env2)

    return



import os
import re

import astropy.units as u
import numpy as np
from astropy.io import fits
from astropy.table import Table


def make_directories(output_dir, fnames, line_names,
                     run_drw_rej, run_pyccf, run_pyzdcf, run_pyroa,
                     reject_data, together_pyroa):

    #Create subdirectories for each line
    for i in range(len(fnames)):
        os.makedirs( output_dir + line_names[i], exist_ok=True )

        if (run_drw_rej) & (reject_data[i]):
            os.makedirs( output_dir + line_names[i] + '/drw_rej', exist_ok=True )

    for i in range(len(fnames)-1):
        if run_pyccf:
            os.makedirs( output_dir + line_names[i+1] + '/pyccf', exist_ok=True )
        if run_pyzdcf:
            os.makedirs( output_dir + line_names[i+1] + '/pyzdcf', exist_ok=True )

    if run_pyroa:
        if together_pyroa:
            os.makedirs( output_dir + 'pyroa', exist_ok=True )
        else:
            for i in range(len(fnames)-1):
                os.makedirs( output_dir + line_names[i+1] + '/pyroa', exist_ok=True )


    #Make subdirectories for light curves
    os.makedirs( output_dir + 'light_curves/', exist_ok=True )

    return






def str2unit(x):

    if (x == 'Arbitrary Units'):
        unit = u.dimensionless_unscaled
    else:
        unit = u.Unit(x)

    return unit


#From https://stackoverflow.com/questions/5807952/removing-trailing-zeros-in-python/5808661#5808661
def number_shaver(ch,
                  regx = re.compile('(?<![\d.])0*(?:'
                                    '(\d+)\.?|\.(0)'
                                    '|(\.\d+?)|(\d+\.\d+?)'
                                    ')0*(?![\d.])')  ,
                  repl = lambda mat: mat.group(mat.lastindex)
                                     if mat.lastindex!=3
                                     else '0' + mat.group(3) ):
    return regx.sub(repl,ch)


def err2str(val, up_err, lo_err, dec=2):
    val = number_shaver( str(round(val, dec)) )
    up_err = number_shaver( str(round(up_err, dec)) )
    lo_err = number_shaver( str(round(lo_err, dec)) )

    return val + '^{+' + up_err + '}_{-' + lo_err + '}'


def write_data(arr, fname, header=None):
    arr = np.array(arr, dtype=object)

    ndim = len( arr.shape )
    assert ndim <= 2


    if ndim == 2:
        cols = len(arr)
        rows = len(arr[0])

        with open(fname, "w") as file:

            if header is not None:
                file.write(header + "\n")

            for i in range(rows):

                string = "{},".format(arr[0][i])
                for j in range(1, cols-1):
                    string += "{},".format(arr[j][i])

                string += "{}\n".format(arr[-1][i])

                file.write(string)

    elif ndim == 1:
        rows = len(arr)

        with open(fname, "w") as file:

            if header is not None:
                file.write(header + "\n")

            for i in range(rows):
                file.write( "{}\n".format(arr[i]) )

    return



def write_weighting_summary(fname, res, run_pyccf, run_javelin, run_pyroa):

    run_arr = [run_pyccf, run_javelin, run_pyroa]
    names = ['pyccf', 'javelin', 'pyroa']
    mod_ncols = 6

    colnames = ['n0', 'peak_bounds', 'peak', 'lag', 'lag_err', 'frac_rejected']
    colnames_tot = []
    for name in names:
        for col in colnames:
            colnames_tot.append( col + '_' + name )

    weighting_dat = []
    weighting_dat.append( res['k'] )
    for i, run in enumerate(run_arr):
        if run:
            for i in range( i*mod_ncols, (i+1)*mod_ncols ):
                weighting_dat.append( res[ colnames_tot[i] ] )

        else:
            for i in range(mod_ncols):
                if i in [1, 4]:
                    weighting_dat.append( [np.nan, np.nan] )
                else:
                    weighting_dat.append( np.nan )

    if run_pyccf:
        weighting_dat.append( res['rmax_pyccf'] )

        if run_javelin & run_pyroa:
            weighting_dat.append( res['rmax_javelin'] )
            weighting_dat.append( res['rmax_pyroa'] )
        elif run_javelin:
            weighting_dat.append( res['rmax_javelin'] )
            weighting_dat.append( np.nan )
        elif run_pyroa:
            weighting_dat.append( np.nan )
            weighting_dat.append( res['rmax_pyroa'] )
        else:
            weighting_dat.append( np.nan )
            weighting_dat.append( np.nan )

    else:
        weighting_dat.append(np.nan)
        weighting_dat.append(np.nan)
        weighting_dat.append(np.nan)


    tot_fits_cols = []

    fits_colnames = []
    fits_colnames.append('k')
    for i in range(len(colnames_tot)):
        fits_colnames.append( colnames_tot[i] )

    fits_colnames.append('rmax_pyccf')
    fits_colnames.append('rmax_javelin')
    fits_colnames.append('rmax_pyroa')



    col_fmts = []
    col_fmts.append('E')
    for _ in range(len(run_arr)):
        for j in range(mod_ncols):
            if j in [1,4]:
                col_fmts.append('2E')
            else:
                col_fmts.append('E')

    for _ in range(3):
        col_fmts.append('E')


    for i in range(len(weighting_dat)):
        fits_col_i = fits.Column(name=fits_colnames[i], format=col_fmts[i], array=[ weighting_dat[i] ])
        tot_fits_cols.append(fits_col_i)

    table = fits.BinTableHDU.from_columns(tot_fits_cols)
    table.writeto(fname, overwrite=True)

    return


def combine_weight_summary(filenames, output_fname, line_names=None):

    if line_names is None:
        line_names = []

        for i in range(len(filenames)):
            line_names.append('Line {}'.format(i+1))


    #---------------------------
    #Get header
    hdr = fits.open(filenames[0]).header

    #---------------------------
    #Get data

    k = np.zeros(len(filenames))

    n0_pyccf = np.zeros(len(filenames))
    peak_bounds_pyccf = np.zeros( ( len(filenames),2 ) )
    peak_pyccf = np.zeros(len(filenames))
    lag_pyccf = np.zeros(len(filenames))
    lag_err_pyccf = np.zeros( ( len(filenames),2 ) )
    frac_rejected_pyccf = np.zeros(len(filenames))

    n0_javelin = np.zeros(len(filenames))
    peak_bounds_javelin = np.zeros( (len(filenames),2) )
    peak_javelin = np.zeros(len(filenames))
    lag_javelin = np.zeros(len(filenames))
    lag_err_javelin = np.zeros( (len(filenames),2) )
    frac_rejected_javelin = np.zeros(len(filenames))

    n0_pyroa = np.zeros(len(filenames))
    peak_bounds_pyroa = np.zeros( (len(filenames),2) )
    peak_pyroa = np.zeros(len(filenames))
    lag_pyroa = np.zeros(len(filenames))
    lag_err_pyroa = np.zeros( (len(filenames),2) )
    frac_rejected_pyroa = np.zeros(len(filenames))

    rmax_pyccf = np.zeros(len(filenames))
    rmax_jav = np.zeros(len(filenames))
    rmax_pyroa = np.zeros(len(filenames))


    for i, fname in enumerate(filenames):
        table = Table.read(fname)

        k[i] = table['k'][0]

        n0_pyccf[i] = table['n0_pyccf'][0]
        peak_bounds_pyccf[i,:] = table['peak_bounds_pyccf'][0]
        peak_pyccf[i] = table['peak_pyccf'][0]
        lag_pyccf[i] = table['lag_pyccf'][0]
        lag_err_pyccf[i,:] = table['lag_err_pyccf'][0]
        frac_rejected_pyccf[i] = table['frac_rejected_pyccf'][0]

        n0_javelin[i] = table['n0_javelin'][0]
        peak_bounds_javelin[i,:] = table['peak_bounds_javelin'][0]
        peak_javelin[i] = table['peak_javelin'][0]
        lag_javelin[i] = table['lag_javelin'][0]
        lag_err_javelin[i,:] = table['lag_err_javelin'][0]
        frac_rejected_javelin[i] = table['frac_rejected_javelin'][0]

        n0_pyroa[i] = table['n0_pyroa'][0]
        peak_bounds_pyroa[i,:] = table['peak_bounds_pyroa'][0]
        peak_pyroa[i] = table['peak_pyroa'][0]
        lag_pyroa[i] = table['lag_pyroa'][0]
        lag_err_pyroa[i,:] = table['lag_err_pyroa'][0]
        frac_rejected_pyroa[i] = table['frac_rejected_pyroa'][0]

        rmax_pyccf[i] = table['rmax_pyccf'][0]
        rmax_jav[i] = table['rmax_javelin'][0]
        rmax_pyroa[i] = table['rmax_pyroa'][0]

    #---------------------------
    #Make columns

    name_col = fits.Column(name='name', format='20A', array=line_names)
    k_col = fits.Column(name='k', format='E', array=k)

    n0_pyccf_col = fits.Column(name='n0_pyccf', format='E', array=n0_pyccf)
    peak_bounds_pyccf_col = fits.Column(name='peak_bounds_pyccf', format='2E', array=peak_bounds_pyccf)
    peak_pyccf_col = fits.Column(name='peak_pyccf', format='E', array=peak_pyccf)
    lag_pyccf_col = fits.Column(name='lag_pyccf', format='E', array=lag_pyccf)
    lag_err_pyccf_col = fits.Column(name='lag_err_pyccf', format='2E', array=lag_err_pyccf)
    frac_rejected_pyccf_col = fits.Column(name='frac_rejected_pyccf', format='E', array=frac_rejected_pyccf)

    n0_javelin_col = fits.Column(name='n0_javelin', format='E', array=n0_javelin)
    peak_bounds_javelin_col = fits.Column(name='peak_bounds_javelin', format='2E', array=peak_bounds_javelin)
    peak_javelin_col = fits.Column(name='peak_javelin', format='E', array=peak_javelin)
    lag_javelin_col = fits.Column(name='lag_javelin', format='E', array=lag_javelin)
    lag_err_javelin_col = fits.Column(name='lag_err_javelin', format='2E', array=lag_err_javelin)
    frac_rejected_javelin_col = fits.Column(name='frac_rejected_javelin', format='E', array=frac_rejected_javelin)

    n0_pyroa_col = fits.Column(name='n0_pyroa', format='E', array=n0_pyroa)
    peak_bounds_pyroa_col = fits.Column(name='peak_bounds_pyroa', format='2E', array=peak_bounds_pyroa)
    peak_pyroa_col = fits.Column(name='peak_pyroa', format='E', array=peak_pyroa)
    lag_pyroa_col = fits.Column(name='lag_pyroa', format='E', array=lag_pyroa)
    lag_err_pyroa_col = fits.Column(name='lag_err_pyroa', format='2E', array=lag_err_pyroa)
    frac_rejected_pyroa_col = fits.Column(name='frac_rejected_pyroa', format='E', array=frac_rejected_pyroa)

    rmax_pyccf_col = fits.Column(name='rmax_pyccf', format='E', array=rmax_pyccf)
    rmax_jav_col = fits.Column(name='rmax_javelin', format='E', array=rmax_jav)
    rmax_pyroa_col = fits.Column(name='rmax_pyroa', format='E', array=rmax_pyroa)

    cols = [name_col, k_col,
            n0_pyccf_col, peak_bounds_pyccf_col, peak_pyccf_col, lag_pyccf_col, lag_err_pyccf_col, frac_rejected_pyccf_col,
            n0_javelin_col, peak_bounds_javelin_col, peak_javelin_col, lag_javelin_col, lag_err_javelin_col, frac_rejected_javelin_col,
            n0_pyroa_col, peak_bounds_pyroa_col, peak_pyroa_col, lag_pyroa_col, lag_err_pyroa_col, frac_rejected_pyroa_col,
            rmax_pyccf_col, rmax_jav_col, rmax_pyroa_col]

    #---------------------------
    #Make table

    table_hdu = fits.BinTableHDU.from_columns(cols, header=hdr)
    table_hdu.writeto(output_fname, overwrite=True)

    return



import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import gridspec
from mpl_toolkits.axes_grid1.inset_locator import (BboxConnector,
                                                   TransformedBbox, inset_axes)

from pypetal.utils.petalio import err2str

mpl.rcParams['xtick.minor.visible'] = True
mpl.rcParams['xtick.top'] = True
mpl.rcParams['xtick.direction'] = 'in'
mpl.rcParams['xtick.major.size'] = 7
mpl.rcParams['xtick.major.size'] = 3

mpl.rcParams['ytick.minor.visible'] = True
mpl.rcParams['ytick.right'] = True
mpl.rcParams['ytick.direction'] = 'in'
mpl.rcParams['ytick.major.size'] = 7
mpl.rcParams['ytick.major.size'] = 3

mpl.rcParams["figure.autolayout"] = False

mpl.rcParams['savefig.dpi'] = 200
mpl.rcParams['savefig.format'] = 'pdf'



def plot_weights(output_dir, line_name, res, n0, k, time_unit='d', plot=False):

    """Plot the weight distribution for a given object for a given module's output distribution.

    Parameters
    ----------

    output_dir : str
        Directory to save plots to.

    line_name : str
        Name of the line.

    res : dict
        Dictionary containing the results from the module.

    n0 : int
        The number of overlapping data points at a lag tau = 0.

    k : int, optional
        The power the probability distribution was raised to.

    time_unit : str, optional
        The time unit of the lag distribution. Default is 'd'.

    plot : bool, optional
        Whether to show the plotted results. Default is ``False``.



    Returns
    -------

    fig : matplotlib.figure.Figure
        The figure object containing the plot.

    ax : matplotlib.axes.Axes
        The axes object for the plot.

    """


    lags = res['lags'][-1].copy()
    acf = res['acf'][-1].copy()
    prob_dist = res['weight_dist'][-1].copy()
    ntau = res['ntau'][-1].copy()

    #Get P(tau)
    probs = (ntau/n0)**k

    #Set all ACF values past first negative to zero
        #Left side
    left_inds = np.argwhere( (acf < 0) & (lags < 0) ).T[0]
    if len(left_inds) == 0:
        ind1 = 0
    else:
        ind1 = np.max( left_inds )

        #Right side
    right_inds = np.argwhere( (acf < 0) & (lags > 0) ).T[0]
    if len(right_inds) == 0:
        ind2 = len(acf)-1
    else:
        ind2 = np.min( right_inds )

    acf_new = acf.copy()
    acf_new[:ind1+1] = 0
    acf_new[ind2:] = 0



    fig, ax = plt.subplots()

    ax.plot(lags, probs, c='gray', label=r'P($\tau$) = $[N(\tau) \ / \ N(0)]^{' + '{}'.format(k) + '}$')
    ax.plot(lags, acf_new, c='r', lw=1.5, label='ACF')
    ax.plot( lags, prob_dist/np.max(prob_dist), c='b', lw=2, label = r'P($\tau$) * ACF' )

    ax.set_xlabel('Lag [' + str(time_unit) + ']', fontsize=16)
    ax.set_ylabel('Weights', fontsize=16)

    ax.tick_params(labelsize=11)
    ax.tick_params('both', which='major', length=8)
    ax.tick_params('both', which='minor', length=4)

    plt.figlegend(bbox_to_anchor=(1.3,.9), fontsize=12)
    plt.savefig( output_dir + line_name + r'/weights/' + line_name + '_weights.pdf', dpi=200, bbox_inches='tight' )

    if plot:
        plt.show()

    plt.cla()
    plt.clf()
    plt.close()

    return fig, ax



def plot_weight_output(output_dir, cont_fname, line_fnames, line_names,
                 res, general_kwargs, module, zoom=False, plot=True):

    """Plot the output of the `run_weights` function. Will output histograms for each line
    of the given time lag distribution for the module, after using the weighting procedure
    described by Grier et al. (2019).
    The plot will have a panel for each line. Each panel will contain three subplots within them.
    The top plot will have the weighting function w($\tau$), the ACF of the continuum, and the smoothed
    distribution used to find the peak. The main panel will have the original distribution and the weighted
    distribution after applying w($\tau$). Additionally, there can be an inset plot zooming in on
    the peak of the distribution (see ``zoom``).


    Parameters
    ----------
    output_dir : str
        The directory to output the plots to.
    cont_fname : str
        The name of the continuum file.
    line_fnames : list of str
        The names of the line files.
    line_names : list of str
        The names of the light curves.
    res : dict
        The output of the `run_weights` function for a given module. For example, if ``res``
        is the output of ``run_weights`` when ``run_pyccf=True``, then this input should be
        ``res['pyccf']`` to plot the pyCCF results.
    general_kwargs : dict
        The general keyword arguments used in ``run_pipeline`` function.
    module : str
        The name of the module this function is plotting the results for. Should be either
        'pyccf' or 'javelin'.
    zoom : bool, optional
        If ``True``, the plots will be zoomed in on the peak of the distribution. Default is
        ``False``.
    plot : bool, optional
        If ``True``, the plots will be displayed. Default is ``True``.
    .. note:: Even if ``zoom=False``, the plots will show an inset on the peak of the distribution if the range of the distribution is too large.
    """


    #Read general kwargs
    plot = general_kwargs['plot']
    time_unit = general_kwargs['time_unit']



    nlc = len(res['lags'])

    Ncol = 3
    Nrow = nlc//Ncol + 1

    #--------------------------------------------------------------------------------

    if nlc <= 3:
        Nrow = 1
        Ncol = nlc


    fig = plt.figure(figsize=( 6*Ncol, 5*Nrow ))

    gs = gridspec.GridSpec(Nrow, Ncol, figure=fig)
    ax_tot = np.zeros( (Nrow, Ncol), dtype=object )

    for i in range(Nrow):
        for j in range(Ncol):
            sub_gs = gridspec.GridSpecFromSubplotSpec(2, 1, subplot_spec=gs[i, j],
                                                    height_ratios=[1, 3], hspace=0)

            ax_bot = plt.subplot(sub_gs[1])
            ax_top = plt.subplot(sub_gs[0], sharex=ax_bot)

            ax_tot[i, j] = [ax_top, ax_bot]


    for i in range(nlc):

        if module == 'pyccf':
            lags = res['lags'][i].copy()
            lag_dist = res['CCCD'][i].copy()

            lag_err_lo = res['centroid'][i][0]
            lag_err_hi = res['centroid'][i][2]
            lag_value = res['centroid'][i][1]

            xlabel = r'{ \rm CCCD }'

        elif module == 'javelin':
            lags = res['lags'][i].copy()
            lag_dist = res['lag_dist'][i].copy()

            lag_err_lo = res['tophat_lag'][i][0]
            lag_err_hi = res['tophat_lag'][i][2]
            lag_value = res['tophat_lag'][i][1]

            xlabel = r't'

        elif module == 'pyroa':
            lags = res['lags'][i].copy()
            lag_dist = res['lag_dist'][i].copy()

            lag_err_lo = res['time_delay'][i][0]
            lag_err_hi = res['time_delay'][i][2]
            lag_value = res['time_delay'][i][1]

            xlabel = r'\tau'


        llim = res['bounds'][i][0]
        rlim = res['bounds'][i][2]
        peak = res['bounds'][i][1]

        weight_dist = res['weight_dist'][i].copy()
        smooth_dist = res['smoothed_dist'][i].copy()
        acf = res['acf'][i].copy()


        #Set ACF=0 when ACF<0
            #Left side
        left_inds = np.argwhere( (acf < 0) & (lags < 0) ).T[0]
        if len(left_inds) == 0:
            ind1 = 0
        else:
            ind1 = np.max( left_inds )

            #Right side
        right_inds = np.argwhere( (acf < 0) & (lags > 0) ).T[0]
        if len(right_inds) == 0:
            ind2 = len(acf)-1
        else:
            ind2 = np.min( right_inds )

        acf[:ind1] = 0
        acf[ind2:] = 0



        if nlc == 1:
            col_ind = 0
            row_ind = 0
        else:
            col_ind = i%Ncol
            row_ind = i//Ncol


        #Plot original lag distribution
        dbin = lags[1] - lags[0]
        bin0 = np.min(lags)

        bin_edges = []
        bin0 = np.min(lags) - dbin/2

        for j in range(len(lags)+1):
            bin_edges.append( bin0 + j*dbin )

        hist, _ = np.histogram(lag_dist, bins=bin_edges)
        hist = np.array(hist, dtype=float)


        #Set top of plot
        good_ind = np.argwhere( ( lags >= llim ) & ( lags <= rlim ) ).T[0]

        if np.argmax(hist) in good_ind:
            ytop = 1.5*np.max(hist)

        elif np.percentile(hist, 99) > np.max(hist):
            ytop = np.percentile(hist, 99)

        else:
            ytop = 1.5*np.max(hist[good_ind])

        #Inset axis?
        lagmax = np.max(lags)
        lagmin = np.min(lags)

        if ( (rlim-llim) < .1*(lagmax-lagmin) ) | zoom:

            #Put inset on the right if peak is on the left and vice-versa
            if peak < (lagmin+lagmax)/2:
                loc=1
            else:
                loc=2

            axins = inset_axes(ax_tot[row_ind, col_ind][1], width='45%', height='40%', loc=loc)

            #Make connectors between the inset and the main plot
            good_ind = np.argwhere( (lags >= llim) & (lags <= rlim) ).T[0]
            if np.max(hist[good_ind]) < .4*ytop:
                loc11 = 3
                loc21 = 2

                loc12 = 4
                loc22 = 1

            else:
                loc11 = 3
                loc21 = 3

                loc12 = 4
                loc22 = 4

            rect = TransformedBbox(axins.viewLim, ax_tot[row_ind, col_ind][1].transData)

            p1 = BboxConnector(axins.bbox, rect, loc1=loc11, loc2=loc21, ec='k')
            axins.add_patch(p1)
            p1.set_clip_on(False)

            p2 = BboxConnector(axins.bbox, rect, loc1=loc12, loc2=loc22, ec='k')
            axins.add_patch(p2)
            p2.set_clip_on(False)

            #Plot the histograms on both the main and inset axes
            for a in [ax_tot[row_ind, col_ind][1], axins]:
                bin1 = a.fill_between(lags, np.zeros_like(hist), hist, step='mid')

                bin2 = a.fill_between(lags, np.zeros_like(hist), weight_dist*hist, step='mid', color='r', alpha=.7)

                a.axvspan( llim, rlim, color='k', alpha=.1 )
                a.set_ylim(0, ytop)

            axins.set_xlim( llim, rlim )
            axins.set_xticks([])


            #If the peak is small, make box around it in the main plot
            if np.max(hist[good_ind]) < .4*ytop:
                y2 = np.max( hist[good_ind] )
                axins.set_ylim(top=1.05*y2)

                x1, x2 = ax_tot[row_ind, col_ind][1].get_xlim()
                xmin = (llim - x1)/(x2-x1)
                xmax = (rlim - x1)/(x2-x1)

                y1_ax, y2_ax = ax_tot[row_ind, col_ind][1].get_ylim()
                ymax = (1.1*y2 - y1_ax)/(y2_ax-y1_ax)

                ax_tot[row_ind, col_ind][1].axhline( 1.05*y2, xmin, xmax, color='k', lw=1.5 )
                ax_tot[row_ind, col_ind][1].axvline( llim, 0, ymax, color='k', lw=1.5 )
                ax_tot[row_ind, col_ind][1].axvline( rlim, 0, ymax, color='k', lw=1.5 )


                yticks = ax_tot[row_ind, col_ind][1].get_yticks()
                m_yticks = ax_tot[row_ind, col_ind][1].get_yticks(minor=True)

                good_ind = np.argwhere( yticks <= 1.05*y2 ).T[0]
                axins.set_yticks( yticks[good_ind] )

                good_ind = np.argwhere( m_yticks <= 1.05*y2 ).T[0]
                axins.set_yticks( m_yticks[good_ind], minor=True )

                axins.tick_params('both', which='major', length=7)
                axins.tick_params('both', which='minor', length=4)

            else:
                y1, y2 = ax_tot[row_ind, col_ind][1].get_ylim()
                axins.set_ylim(y1, y2)

                yticks = ax_tot[row_ind, col_ind][1].get_yticks()
                m_yticks = ax_tot[row_ind, col_ind][1].get_yticks(minor=True)

                axins.set_yticks( yticks )
                axins.set_yticks( m_yticks, minor=True )

                axins.tick_params('both', which='major', length=0)
                axins.tick_params('both', which='minor', length=0)


            axins.set_yticklabels([])



        else:
            loc=1

            bin1 = ax_tot[row_ind, col_ind][1].fill_between(lags, np.zeros_like(hist), hist, step='mid')
            bin2 = ax_tot[row_ind, col_ind][1].fill_between(lags, np.zeros_like(hist), weight_dist*hist, step='mid', color='r', alpha=.7)
            ax_tot[row_ind, col_ind][1].axvspan( llim, rlim, color='k', alpha=.1 )

            ax_tot[row_ind, col_ind][1].set_ylim(0, ytop)


        #Plot ACF
        im3, = ax_tot[row_ind, col_ind][0].plot(lags, acf, c='r')


        #Plot weighting function
        im1, = ax_tot[row_ind, col_ind][0].plot(lags, weight_dist, c='k')
        ax_tot[row_ind, col_ind][0].set_yticks([.5, 1])

        #Plot smoothed distribution
        im2, = ax_tot[row_ind, col_ind][0].plot(lags, smooth_dist/np.max(smooth_dist), c='DodgerBlue')
        ax_tot[row_ind, col_ind][0].axvspan( llim, rlim, color='k', alpha=.1 )

        #Write lag and error
        peak_str = err2str( lag_value, lag_err_hi, lag_err_lo, dec=2 )
        peak_str = r'$' + r'{}'.format(peak_str) + r'$' + ' ' + time_unit

        if loc == 1:
            xtxt = .05
            ha='left'
        else:
            xtxt = .95
            ha='right'

        ax_tot[row_ind, col_ind][1].text( xtxt, .85, peak_str,
                                    ha=ha, transform=ax_tot[row_ind, col_ind][1].transAxes,
                                    fontsize=13 )

        ax_tot[row_ind, col_ind][1].set_xlabel( r'$' + xlabel + r'_{' + r'{}'.format(line_names[i+1]) + r'}$ [' + time_unit + ']', fontsize=15 )


    for i in range(Nrow):
        ax_tot[i, 0][1].set_ylabel('N', fontsize=16)

    for i in range(Nrow):
        for j in range(Ncol):
            ax_tot[i,j][1].tick_params('both', labelsize=11)
            ax_tot[i,j][1].tick_params('both', which='major', length=7)
            ax_tot[i,j][1].tick_params('both', which='minor', length=3)

            ax_tot[i,j][0].tick_params('y', labelsize=11)
            ax_tot[i,j][0].tick_params('x', labelsize=0)
            ax_tot[i,j][0].tick_params('both', which='major', length=7)
            ax_tot[i,j][0].tick_params('both', which='minor', length=3)


    plt.subplots_adjust(wspace=.15)

    #Put legends for the top and bottom plots
    for i in range(Nrow):
        ax_tot[i, -1][0].legend( [im1, im2, im3], [r'w($\tau$)', 'Smoothed Dist', 'ACF'],
                                bbox_to_anchor=(1,1.1), fontsize=11, loc='upper left' )

        ax_tot[i, -1][1].legend( [bin1, bin2], ['Original', 'Weighted'],
                                bbox_to_anchor=(1, 1), fontsize=11, loc='upper left')


    plt.savefig( output_dir + module + '_weights_res.pdf', dpi=200, bbox_inches='tight' )


    if plot:
        plt.show()

    plt.cla()
    plt.clf()
    plt.close()

    return



import corner
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import palettable
import scipy
from matplotlib import gridspec
from matplotlib.colors import ListedColormap
from PyROA.PyROA import (Gaussian, InverseGaussian, LogGaussian, TruncGaussian,
                         Uniform)

from pypetal.pyroa.utils import get_samples_chunks

mpl.rcParams['xtick.minor.visible'] = True
mpl.rcParams['xtick.top'] = True
mpl.rcParams['xtick.direction'] = 'in'
mpl.rcParams['xtick.major.size'] = 7
mpl.rcParams['xtick.major.size'] = 3

mpl.rcParams['ytick.minor.visible'] = True
mpl.rcParams['ytick.right'] = True
mpl.rcParams['ytick.direction'] = 'in'
mpl.rcParams['ytick.major.size'] = 7
mpl.rcParams['ytick.major.size'] = 3

mpl.rcParams["figure.autolayout"] = False

mpl.rcParams['savefig.dpi'] = 200
mpl.rcParams['savefig.format'] = 'pdf'





def plot_histograms(samples, line_names, nburn=0, add_var=False, delay_dist=False, nbin=50, fname=None, show=False):


    """Plot histograms of the posterior distributions of the parameters from PyROA.

    Parameters
    ----------

    samples : list of float
        The samples object output from PyROA. Can be found in the "samples.obj" file.

    line_names: list of str
        The names of the lines corresponding to the samples.

    nburn : int, optional
        The number of burn-in samples to discard. Default is 0.

    add_var : bool, optional
        The add_var parameter from PyROA. Default is False.

    delay_dist : bool, optional
        The delay_dist parameter from PyROA. Default is False.

    nbin : int, optional
        The number of bins to use in the histograms. Default is 50.

    fname : str, optional
        The name of the file to save the plot to. If None, the plot will not be saved. Default is None.

    show : bool, optional
        Whether to show the plot. Default is False.


    Returns
    -------

    """


    samples_chunks = get_samples_chunks(samples, nburn, add_var, delay_dist)

    names = ['A', 'B', r'\tau']
    if delay_dist:
        names.append(r'\Delta')
    if add_var:
        names.append(r'\sigma')


    nlc = len(samples_chunks) - 1
    nrow = nlc + 1
    ncol = 3

    if add_var:
        ncol += 1

    if delay_dist:
        ncol += 1


    fig, ax = plt.subplots(nrow, ncol, figsize=(5*ncol, 5*nrow) )

    for n in range(nlc):
        ax[n, 0].set_ylabel('N', fontsize=20)

        for i in range(len(samples_chunks[0])):
            ax[n,i].hist(samples_chunks[n][i], nbin)

            ax[n,i].tick_params('y', labelsize=13)
            ax[n,i].tick_params('x', labelsize=15)
            ax[n,i].tick_params('both', which='major', length=7)
            ax[n,i].tick_params('both', which='minor', length=3)

            ax[n,i].set_xlabel( r'$' + names[i] + '_{' + line_names[n] + r'}$', fontsize=20)


    #For delta
    for i in range(ncol):
        if i == 0:
            ax[-1, i].set_ylabel('N', fontsize=20)
            ax[-1,i].set_xlabel( r'$\Delta$', fontsize=20 )
            ax[-1,i].hist(samples_chunks[-1][-1], nbin)

            ax[-1,i].tick_params('y', labelsize=13)
            ax[-1,i].tick_params('x', labelsize=15)
            ax[-1,i].tick_params('both', which='major', length=7)
            ax[-1,i].tick_params('both', which='minor', length=3)

        else:
            ax[-1, i].axis('off')


    plt.subplots_adjust(hspace=.3)

    if fname is not None:
        plt.savefig(fname, dpi=200, bbox_inches='tight')

    if show:
        plt.show()

    plt.cla()
    plt.clf()
    plt.close()

    return




def pyroa_trace_plot(samples, line_names, add_var=False, delay_dist=False, nburn=0, fname=None, show=False, alpha=.1):

    """Plot traces of the walkers from the MCMC sampling in PyROA.


    Parameters
    ----------

    samples : list of float
        The samples object output from PyROA. Can be found in the "samples.obj" file.

    line_names: list of str
        The names of the lines corresponding to the samples.

    add_var : bool, optional
        The add_var parameter from PyROA. Default is False.

    delay_dist : bool, optional
        The delay_dist parameter from PyROA. Default is False.

    nburn : int, optional
        The number of burn-in samples to discard. Default is 0.

    fname : str, optional
        The name of the file to save the plot to. If None, the plot will not be saved. Default is None.

    show : bool, optional
        Whether to show the plot. Default is False.

    alpha : float, optional
        The alpha value for the walker traces. Default is 0.1.


    Returns
    -------

    """


    row_labels = [r'$A_i$', r'$B_i$', r'$\tau_i$']
    if delay_dist:
        row_labels.append( r'$\Delta_i$' )
    if add_var:
        row_labels.append( r'$\sigma_i$' )

    nrow = 3
    if add_var:
        nrow += 1
    if delay_dist:
        nrow += 1

    ncol = (samples.shape[2]-3)//nrow + 2
    nparam = samples.shape[2] - 1

    fig, ax = plt.subplots(nrow, ncol, figsize=(5*ncol, nrow), sharex=True)


    #Labels
    for i in range(nrow):
        ax[i,0].set_ylabel(row_labels[i])

    #Do continuum
    for j in range(samples.shape[1]):
        ax[0,0].plot( samples[:,j,0], c='k', alpha=alpha )
        ax[1,0].plot( samples[:,j,1], c='k', alpha=alpha )
        ax[2,0].plot( np.zeros( samples.shape[0] ), c='k', alpha=alpha )
        if add_var & delay_dist:
            ax[3,0].plot( np.zeros( samples.shape[0] ), c='k', alpha=alpha )
            ax[4,0].plot( samples[:,j,2], c='k', alpha=alpha )
        elif add_var:
            ax[3,0].plot( samples[:,j,2], c='k', alpha=alpha )
        elif delay_dist:
            ax[3,0].plot( np.zeros( samples.shape[0] ), c='k', alpha=alpha )

    if nburn > 0:
        for i in range(nrow):
            ax[i,0].axvline(nburn, ls='--', color='r')

    ax[0,0].set_title(line_names[0])


    #Do lines
    if add_var:
        min_i = 3
    else:
        min_i = 2

    for i in range(min_i, nparam):

        if add_var:
            m = (i-min_i)%nrow
            n = (i-min_i)//nrow + 1
        else:
            m = (i-min_i)%nrow
            n = (i-min_i)//nrow + 1


        for j in range(samples.shape[1]):
            ax[m,n].plot( samples[:,j,i], c='k', alpha=alpha )

        if nburn > 0:
            ax[m,n].axvline(nburn, ls='--', color='r')

        if m == 0:
            ax[m,n].set_title(line_names[n])



    #Delta
    for j in range(samples.shape[1]):
        ax[0,-1].plot( samples[:,j,-1], c='k', alpha=alpha )

    if nburn > 0:
        ax[0,-1].axvline(nburn, ls='--', color='r')

    ax[0,-1].set_title(r'$\Delta$')

    for i in range(1, nrow):
        ax[i,-1].axis('off')


    plt.subplots_adjust(wspace=.2)

    if fname is not None:
        plt.savefig(fname, dpi=200, bbox_inches='tight')

    if show:
        plt.show()

    plt.cla()
    plt.clf()
    plt.close()

    return




def peaktomean(mu, a, rms):
    alpha = (a-mu)/rms
    phi = (1.0/np.sqrt(2.0*np.pi))*np.exp(-0.5*alpha**2)
    Z = 1 - 0.5*(1+scipy.special.erf(alpha/np.sqrt(2.0)))
    return mu + rms*phi/Z



def plot_fits(fnames, line_names, samples, models,
              nburn=0, add_var=False, delay_dist=False, psi_types='Gaussian',
              nbin=50, delimiter=',',
              time_unit='d', lc_unit='mJy',
              output_fname=None, show=False):

    if isinstance(psi_types, str):
        psi_types = [psi_types]*( len(fnames)-1 )

    if isinstance(lc_unit, str):
        lc_unit = [lc_unit]*len(fnames)

    #Get colors
    cmap = ListedColormap( palettable.cartocolors.qualitative.Vivid_10.mpl_colors )
    colors = cmap.colors

    #Get samples
    samples_chunks = get_samples_chunks(samples, nburn, add_var, delay_dist)


    nrow = len(fnames)
    fig = plt.figure( figsize=(12, 2*nrow) )

    gs = fig.add_gridspec( nrow, 2, hspace=0, wspace=0, width_ratios=[4,1] )
    ax = gs.subplots(sharex='col')

    for i in range(nrow):
        x, y, yerr = np.loadtxt( fnames[i], usecols=[0,1,2], unpack=True, delimiter=delimiter )
        x_sim, y_sim, yerr_sim = models[i]
        tau_samp = samples_chunks[i][2]

        if delay_dist:
            tau_rms_samp = samples_chunks[i][3]

        if add_var:
            if delay_dist:
                sig_med = np.median(samples_chunks[i][4])
            else:
                sig_med = np.median(samples_chunks[i][3])

            yerr = np.sqrt( yerr**2 + sig_med**2 )


        _, _, bars = ax[i,0].errorbar(x, y, yerr=yerr, fmt='.k', mec='k', mfc=colors[9], zorder=0)
        [bar.set_alpha(.3) for bar in bars]

        ax[i,0].plot(x_sim, y_sim, color=colors[i%9], zorder=-1)
        ax[i,0].fill_between(x_sim, y_sim-yerr_sim, y_sim+yerr_sim, color=colors[i%9], alpha=.3, zorder=-2)


        if i == 0:
            ax[i,1].axis('off')

        if i != 0:

            if delay_dist:

                gs00 = gridspec.GridSpecFromSubplotSpec(2, 1, subplot_spec=ax[i,1], height_ratios=[1, 3], hspace=0, wspace=0 )
                ax1 = fig.add_subplot(gs00[0], sharex=ax[i,1])
                ax2 = fig.add_subplot(gs00[1], sharex=ax[i,1])

                tau = np.median(tau_samp)
                tau_rms = np.median(tau_rms_samp)

                smpls = tau_samp
                if psi_types[i-1] == 'TruncGaussian':
                    smpls = peaktomean( tau_samp, np.zeros_like(tau_samp), tau_rms_samp )

                frq, edges = np.histogram(smpls, bins=nbin)
                ax2.bar( edges[:-1], frq, width=np.diff(edges), align='edge', color=colors[i%9], edgecolor=colors[i%9])



                length = 10*tau_rms
                taus = np.arange( tau - 5*tau_rms, tau + 5*tau_rms, length/500 )
                if (psi_types[i-1] == 'LogGaussian') or (psi_types[i-1] == 'InverseGaussian'):
                    taus = np.linspace( -tau_rms, np.median(samples_chunks[-2][2]) + 5*tau_rms, 600 )

                hi_vals = []
                lo_vals = []
                med_vals = []
                for k in range(len(taus)):
                    if psi_types[i-1] == 'Gaussian':
                        psi = Gaussian( tau_samp, tau_rms_samp, taus[k], conv=False )
                    elif psi_types[i-1] == 'Uniform':
                        psi = Uniform( tau_samp, tau_rms_samp, taus[k], conv=False )
                    elif psi_types[i-1] == 'TruncGaussian':
                        psi = TruncGaussian( tau_samp, tau_rms_samp, taus[k], np.zeros_like(tau_samp), conv=False )
                    elif psi_types[i-1] == 'LogGaussian':
                        psi = LogGaussian( tau_samp, tau_rms_samp, taus[k], np.zeros_like(tau_samp), conv=False )
                    elif psi_types[i-1] == 'InverseGauss':
                        psi = InverseGaussian( tau_samp, tau_rms_samp, taus[k], np.zeros_like(tau_samp), conv=False )

                    hi_vals.append( np.percentile(psi, 84) )
                    lo_vals.append( np.percentile(psi, 16) )
                    med_vals.append( np.median(psi) )


                # norm = np.max(hi_vals)/np.max(frq)

                ax1.plot(taus, med_vals, color='k', zorder=1)
                ax1.fill_between(taus, lo_vals, hi_vals, color='k', alpha=.2, edgecolor='none', zorder=0)
                ax1.set_ylim(0, np.max(med_vals)*1.5)

                ax2.axvline( np.median(tau_samp), ls='--', color='k', lw=.75 )

                for a in [ax1, ax2]:
                    a.set_yticklabels([])
                    a.set_yticks([])

                    a.label_outer()
                    a.tick_params('both', which='major', length=6)
                    a.tick_params('both', which='minor', length=3)

            else:
                ax[i,1].hist(tau_samp, bins=nbin, color=colors[i%9] )
                ax[i,1].axvline( np.median(tau_samp), ls='--', color='k', lw=.75 )


            ax[i,1].set_yticklabels([])
            ax[i,1].set_yticks([])


    tau_min = []
    tau_max = []
    for i in range(1, len(fnames)):
        tau_min.append( np.min(samples_chunks[i][2]) )
        tau_max.append( np.max(samples_chunks[i][2]) )

    length = np.max(tau_max) - np.min(tau_min)
    ax[-1,1].set_xlim( np.min(tau_min) - .25*length, np.max(tau_max) + .25*length )



    ax[-1,0].set_xlabel('Time [{}]'.format(time_unit))
    ax[-1,1].set_xlabel(r'$\tau$ [{}]'.format(time_unit))

    for a in ax.flatten():
        a.label_outer()
        a.tick_params('both', which='major', length=6)
        a.tick_params('both', which='minor', length=3)

    for i in range(nrow):
        if (lc_unit[i] == '') or (lc_unit[i] == 'Arbitrary Unit'):
            ytxt = 'Flux'
        elif lc_unit == 'mag':
            ytxt = 'Magnitude'
            ax[i,0].invert_yaxis()
        else:
            ytxt = 'Flux [{}]'.format(lc_unit[i])


        ax[i,0].text( -.1, .5, ytxt, transform=ax[i,0].transAxes, rotation='vertical', ha='left', va='center' )
        ax[i,0].text( .025, .85, line_names[i], transform=ax[i,0].transAxes, ha='left', va='top' )

        if i != 0:
            tau_med = np.median( samples_chunks[i][2] )
            tau_hi = np.percentile( samples_chunks[i][2], 84)
            tau_lo = np.percentile( samples_chunks[i][2], 16)
            txt = r'$\tau = ' + '{:.2f}'.format(tau_med) + '^{+' + '{:.2f}'.format(tau_hi-tau_med) + '}_{-' + '{:.2f}'.format(tau_med-tau_lo) + '}$'
            ax[i,0].text( .025, .75, txt, transform=ax[i,0].transAxes, ha='left', va='top' )


    if output_fname is not None:
        plt.savefig(output_fname, dpi=200, bbox_inches='tight')

    if show:
       plt.show()

    plt.cla()
    plt.clf()
    plt.close()

    return




def pyroa_corner_plot(samples, line_names, nburn=0, add_var=False, delay_dist=False, split=True, fname=None, show=False):

    if split:

        if fname is not None:
            if isinstance(fname, str) or len(fname) != len(line_names):
                raise Exception('If split=True, a file name must be given for every line')

        samples_chunks = get_samples_chunks(samples, nburn, add_var, delay_dist)

        for i in range(len(line_names)):
            samps_i = []

            if i == 0:
                for j in [0,1]:
                    samps_i.append( samples_chunks[i][j] )
                if add_var:
                    if delay_dist:
                        samps_i.append( samples_chunks[i][4] )
                    else:
                        samps_i.append( samples_chunks[i][3] )

                samps_i.append( samples_chunks[-1][0] )
                samps_i = np.array(samps_i)

                labels = []
                cols = ['A', 'B']
                if add_var:
                    cols.append( r'\sigma' )

                for col in cols:
                    labels.append( r'$' + col + '_{' + line_names[0] + '}$' )
                labels.append( r'$\Delta$' )

                fig = corner.corner( samps_i.T, labels=labels, show_titles=True, plot_contours=True, smooth=1, title_fmt='.3f', label_kwargs={'fontsize':18} );

                if fname is not None:
                    fig.savefig(fname[i], dpi=200, bbox_inches='tight')

                if show:
                    plt.show()

                plt.cla()
                plt.clf()
                plt.close()


            else:
                for j in range(len(samples_chunks[i])):
                    samps_i.append( samples_chunks[i][j] )
                samps_i.append( samples_chunks[-1][0] )
                samps_i = np.array(samps_i)

                labels = []
                cols = ['A', 'B', r'\tau']
                if delay_dist:
                    cols.append( r'\Delta' )
                if add_var:
                    cols.append( r'\sigma' )

                for col in cols:
                    labels.append( r'$' + col + '_{' + line_names[i] + '}$' )
                labels.append( r'$\Delta$' )


                fig = corner.corner( samps_i.T, labels=labels, show_titles=True, plot_contours=True, smooth=1, title_fmt='.3f', label_kwargs={'fontsize':18} );

                if fname is not None:
                    fig.savefig(fname[i], dpi=200, bbox_inches='tight')

                if show:
                    plt.show()

                plt.cla()
                plt.clf()
                plt.close()


    else:
        #Remove burn-in
        samples = samples[nburn:,:,:].copy()
        #Flatten
        samples = samples.reshape( (-1, samples.shape[2]) )


        labels = []
        for col in ['A', 'B']:
            labels.append( r'$' + col + '_{' + line_names[0] + '}$' )
        if add_var:
            labels.append( r'$\sigma_{' + line_names[0] + '}$' )

        cols = ['A', 'B', r'\tau']
        if delay_dist:
            cols.append( r'\Delta' )
        if add_var:
            cols.append( r'\sigma' )

        for name in line_names[1:]:
            for col in cols:
                labels.append( r'$' + col + '_{' + name + '}$' )

        labels.append( r'$\Delta$' )


        fig = corner.corner( samples, labels=labels, show_titles=True, plot_contours=True, smooth=1, title_fmt='.3f', label_kwargs={'fontsize':18} );

        if fname is not None:
            fig.savefig(fname, dpi=200, bbox_inches='tight')

        if show:
            plt.show()

        plt.cla()
        plt.clf()
        plt.close()

    return



import matplotlib as mpl
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt

from pypetal.utils.petalio import err2str

mpl.rcParams['xtick.minor.visible'] = True
mpl.rcParams['xtick.top'] = True
mpl.rcParams['xtick.direction'] = 'in'

mpl.rcParams['ytick.minor.visible'] = True
mpl.rcParams['ytick.right'] = True
mpl.rcParams['ytick.direction'] = 'in'

mpl.rcParams["figure.autolayout"] = False

mpl.rcParams['savefig.dpi'] = 300
mpl.rcParams['savefig.format'] = 'pdf'





def plot_pyzdcf_results(x1, y1, yerr1, x2, y2, yerr2,
                        dcf_df,
                        plike_dict=None,
                        time_unit='d', lc_unit=['mJy', 'mJy'], lc_names=['', ''],
                        fname=None, show=False):

    """Plot the results of pyZDCF.


    Parameters
    ----------

    x1 : list of float
        Time array for the first light curve.

    y1 : list of float
        Values for the first light curve.

    yerr1 : list of float
        Uncertainty in the first light curve.

    x2 : list of float
        Time array for the second light curve.

    y2 : list of float
        Values for the second light curve.

    yerr2 : list of float
        Uncertainty in the second light curve.

    dcf_df : pandas.DataFrame
        The output ``DataFrame`` object from pyZDCF.

    plike_dict : dict, optional
        The output dict from PLIKE, if it is run. Default is ``None``.

    time_unit : str, optional
        The unit of time for the light curves. Default is 'd'.

    lc_unit : str, optional
        The unit of the light curves. Default is 'mJy'.

    lc_names : list of str, optional
        The names of the light curves. Default is ['', ''].

    fname : str, optional
        If not ``None``, will save the plot to the given filename. Default is ``None``.

    show : bool, optional
        If ``True``, will show the plot. Default is ``False``.

    """


    if plike_dict is None:
        include_plike = False
    else:
        include_plike = True

    #Plot the CCF
    fig = plt.figure(  figsize=(12,4))
    gs = gridspec.GridSpec( ncols=3, nrows=2 )


    ax1 = fig.add_subplot( gs[0,:2] )
    ax2 = fig.add_subplot( gs[1,:2], sharex=ax1 )
    ax3 = fig.add_subplot( gs[:,2] )

    lines, caps, bars = ax1.errorbar(x1, y1, yerr1, fmt='.k')
    [bar.set_alpha(.2) for bar in bars]

    ax1.tick_params('x', labelsize=0)


    lines, caps, bars = ax2.errorbar( x2, y2, yerr2, fmt='.k' )
    [bar.set_alpha(.2) for bar in bars]

    ax2.set_xlabel('Time [' + str(time_unit) + ']', fontsize=17)



    lags = dcf_df.tau
    vals = dcf_df.dcf

    err_lo = dcf_df['-err(dcf)']
    err_hi = dcf_df['+err(dcf)']

    tau_err_lo = dcf_df['-sig(tau)']
    tau_err_hi = dcf_df['+sig(tau)']

    lines, caps, bars = ax3.errorbar(lags, vals, yerr=[err_lo, err_hi],
                                     xerr=[tau_err_lo, tau_err_hi], fmt='.k')
    [bar.set_alpha(.2) for bar in bars]


    if include_plike:
        ax3.axvline( plike_dict['ML_lag'], c='b' )
        ax3.axvspan( plike_dict['ML_lag'] - plike_dict['ML_lag_err_lo'],
                    plike_dict['ML_lag'] + plike_dict['ML_lag_err_hi'],
                    color='b', alpha=.1)

        lag_txt = err2str(plike_dict['ML_lag'], plike_dict['ML_lag_err_hi'], plike_dict['ML_lag_err_lo'])
        lag_txt = r'$\tau_{\rm ML} = ' + lag_txt + '$'
        ax2.text( .04, .6, lag_txt, transform=ax2.transAxes, fontsize=15, color='b')



    ax3.set_xlabel('Lag [' + str(time_unit) + ']', fontsize=17)

    plt.figtext(.95, .5, 'ZDCF', fontsize=17, rotation=270, va='center')

    ax3.yaxis.tick_right()
    ax3.yaxis.set_ticks_position('both')

    for i, ax in enumerate([ax1, ax2]):
        y1, y2 = ax.get_ylim()
        y2 = y1 + (y2-y1)*1.3
        ax.set_ylim(y1, y2)

        ax.text( .04, .8, lc_names[i], transform=ax.transAxes, fontsize=15 )


    for ax in [ax1, ax2, ax3]:
        ax.tick_params('both', which='major', length=7)
        ax.tick_params('both', which='minor', length=3)


    for ax in [ax1, ax2]:
        ax.tick_params('y', labelsize=13)
        ax.tick_params('x', labelsize=12)

    ax3.tick_params('both', labelsize=12)
    ax3.locator_params('x', nbins=5)
    ax3.locator_params('y', nbins=5)



    #Add ylabels
    if lc_unit[0] == 'mag':
        ytxt = 'Magnitude'
        ax1.invert_yaxis()

    elif (lc_unit[0] == 'Arbitrary Units') or (lc_unit[0] == ''):
        ytxt = 'Flux'
    else:
        ytxt = 'Flux [' + str(lc_unit[0]) + ']'
    plt.figtext( .06, .68, ytxt, fontsize=16, va='center', rotation=90 )


    if lc_unit[1] == 'mag':
        ytxt = 'Magnitude'
        ax2.invert_yaxis()

    elif (lc_unit[1] == 'Arbitrary Units') or (lc_unit[1] == ''):
        ytxt = 'Flux'
    else:
        ytxt = 'Flux [' + str(lc_unit[1]) + ']'
    plt.figtext( .06, .3, ytxt, fontsize=16, va='center', rotation=90 )

    plt.subplots_adjust(hspace=.03, wspace=.06)

    if fname is not None:
        plt.savefig(fname, dpi=300, bbox_inches='tight')

    if show:
        plt.show()

    plt.cla()
    plt.clf()
    plt.close()

    return



import numpy as np
from astropy.table import Table


def set_general(input_args, fnames):

    default_kwargs = {
        'verbose': False,
        'plot': False,
        'time_unit': 'd',
        'lc_unit': 'Arbitrary Units',
        'file_fmt': 'csv',
        'lag_bounds': 'baseline',
        'threads': 1
    }

    params = { **default_kwargs, **input_args }

    verbose = params['verbose']
    plot = params['plot']
    time_unit = params['time_unit']
    lc_unit = params['lc_unit']
    file_fmt = params['file_fmt']
    lag_bounds = params['lag_bounds']
    threads = params['threads']

    if isinstance(lag_bounds, list):
        if len(lag_bounds) == 1:
            assert isinstance(lag_bounds[0], list) or isinstance(lag_bounds[0], str) or ( lag_bounds[0] is None )

        elif len(lag_bounds) == 2:
            pass

        else:
            for i in range(len(lag_bounds)):
                if isinstance(lag_bounds[i], list):
                    assert ( len(lag_bounds[i]) == 2 ) | ( lag_bounds[i] is None ) | ( lag_bounds[i] == 'baseline' )
                else:
                    assert ( lag_bounds[i] is None ) | ( lag_bounds[i] == 'baseline' )


    if isinstance(lc_unit, str):
        lc_unit = list( np.full( len(fnames), lc_unit ) )




    len_lag_bounds = len(lag_bounds)
    if len_lag_bounds == 2:
        if not ( isinstance(lag_bounds[0], list) or ( lag_bounds[0] == 'baseline' ) ):
            len_lag_bounds = 1
        else:
            len_lag_bounds = 2


    if len_lag_bounds != len(fnames)-1:

        lag_bounds_og = lag_bounds
        lag_bounds = []

        for i in range(len(fnames)-1):
            lag_bounds.append(lag_bounds_og)

    if (len(fnames) == 2) and ( not isinstance(lag_bounds[0], list) ):
        if not isinstance(lag_bounds[0], str):
            lag_bounds = [lag_bounds]



    xvals_tot = []
    baselines = []
    for i in range(len(fnames)):

        try:
            dat = Table.read(fnames[i], format=file_fmt)
        except:
            dat = Table.read(fnames[i], format='ascii')

        colnames = dat.colnames
        x = np.array( dat[colnames[0]] )

        sort_ind = np.argsort(x)
        x = x[sort_ind]

        xvals_tot.append(x)

    for i in range(len(fnames)-1):
        baseline = np.max([ xvals_tot[0].max(), xvals_tot[i+1].max() ]) - np.min([ xvals_tot[0].min(), xvals_tot[i+1].min() ])
        baselines.append(baseline)

    for i in range(len(fnames)-1):
        if (lag_bounds[i] is None) | (lag_bounds[i] == 'baseline'):
            lag_bounds[i] = [-baselines[i], baselines[i]]


    #Return dict so it can be passed to other functions
    output = {
        'verbose': verbose,
        'plot': plot,
        'time_unit': time_unit,
        'lc_unit': lc_unit,
        'file_fmt': file_fmt,
        'lag_bounds': lag_bounds,
        'threads': threads
    }

    return output





def set_drw_rej(input_args, fnames):

    default_kwargs = {
        'jitter' : True,
        'nsig' : 3,
        'nwalker': 100,
        'nburn': 300,
        'nchain': 1000,
        'clip': np.full( len(fnames), True),
        'reject_data': np.hstack([ [True], np.full( len(fnames)-1, False) ]),
        'use_for_javelin': False
    }

    params = { **default_kwargs, **input_args }

    jitter = params['jitter']
    nsig = params['nsig']
    nwalker = params['nwalker']
    nburn = params['nburn']
    nchain = params['nchain']
    clip = params['clip']
    reject_data = params['reject_data']
    use_for_javelin = params['use_for_javelin']

    if isinstance(clip, bool):
        clip = np.full( len(fnames), clip)

    if isinstance(reject_data, bool):
        reject_data = np.full( len(fnames), reject_data)

    if (use_for_javelin) & (not reject_data[0]):
        raise Exception('Cannot use continuum for Javelin without fitting it to a DRW first')


    return jitter, nsig, nwalker, nburn, nchain, clip, \
        reject_data, use_for_javelin


def set_detrend(input_args):

    default_kwargs = {
        'K': 2,
        'nchain': 4,
        'miniter': 5000,
        'maxiter': 10000
    }

    params = { **default_kwargs, **input_args }

    K = params['K']
    nchain = params['nchain']
    miniter = params['miniter']
    maxiter = params['maxiter']

    return K, nchain, miniter, maxiter


def set_pyccf(input_args):

    default_kwargs = {
        'interp': 2 + 1e-10,
        'nsim': 3000,
        'mcmode': 0,
        'sigmode': 0.2,
        'thres': 0.8,
        'nbin': 50
    }

    params = { **default_kwargs, **input_args }

    interp = params['interp']
    nsim = params['nsim']
    mcmode = params['mcmode']
    sigmode = params['sigmode']
    thres = params['thres']
    nbin = params['nbin']

    return interp, nsim, mcmode, sigmode, thres, nbin



def set_pyzdcf(input_args, fnames):

    default_kwargs = {
        'nsim': 1000,
        'minpts': 0,
        'uniform_sampling': False,
        'omit_zero_lags': True,
        'sparse': 'auto',
        'prefix': 'zdcf',
        'run_plike': False,
        'plike_dir': None,
    }

    params = { **default_kwargs, **input_args }

    nsim = params['nsim']
    minpts = params['minpts']
    uniform_sampling = params['uniform_sampling']
    omit_zero_lags = params['omit_zero_lags']
    sparse = params['sparse']
    prefix = params['prefix']
    run_plike = params['run_plike']
    plike_dir = params['plike_dir']


    return nsim, minpts, uniform_sampling, omit_zero_lags, \
        sparse, prefix, run_plike, plike_dir


def set_pyroa(input_args, nlc):

    default_kwargs = {
        'nchain': 20000,
        'nburn': 15000,
        'init_tau': None,
        'subtract_mean': True,
        'div_mean': False,
        'add_var': True,
        'delay_dist': True,
        'psi_types': 'Gaussian',
        'together': True,
        'objname': None,
        'prior_func': None
    }

    params = { **default_kwargs, **input_args }

    nchain = params['nchain']
    nburn = params['nburn']
    init_tau = params['init_tau']
    subtract_mean = params['subtract_mean']
    div_mean = params['div_mean']
    add_var = params['add_var']
    delay_dist = params['delay_dist']
    psi_types = params['psi_types']
    together = params['together']
    objname = params['objname']

    if init_tau is None:
        init_tau = [10.] * (nlc-1)

    if objname is None:
        objname = 'pyroa'

    if isinstance(psi_types, str):
        psi_types = [psi_types] * (nlc-1)

    if not together:
        if isinstance(delay_dist, bool):
            delay_dist = [delay_dist] * (nlc-1)
        if isinstance(add_var, bool):
            add_var = [add_var] * (nlc-1)


    return nchain, nburn, init_tau, subtract_mean, div_mean, \
            add_var, delay_dist, psi_types, together, objname



def set_weighting(input_args):

    default_kwargs = {
        'gap_size': 30,
        'k': 2,
        'width': 15,
        'rel_height': .99,
        'zoom': True
    }

    params = { **default_kwargs, **input_args }

    gap_size = params['gap_size']
    k = params['k']
    width = params['width']
    rel_height = params['rel_height']
    zoom = params['zoom']

    return gap_size, k, width, rel_height, zoom



import os
import subprocess

import numpy as np
from pyzdcf import pyzdcf


def run_plike(dcf_fname, lag_bounds, plike_dir, verbose=False):

    """
    Runs the PLIKE algorithm to compute the maximum likelihood peak of the ZDCF.
    Must have PLIKE (https://ui.adsabs.harvard.edu/abs/2013arXiv1302.1508A/abstract).
    Will output a file containing the PLIKE results ('plike.out') in ``plike_dir``.


    Parameters
    ----------

    plike_dir : str
            Path to the directory with the PLIKE executable.

    lag_bounds : (2,) array_like
            Lower and upper bounds of lags to search for ML peak.

    dcf_name :str
            Path to the ZDCF file, usually ouptput by pyZDCF.

    verbose : bool, optional
            If True, will read output of PLIKE. Default is False.


    Returns
    -------

    """

    #Make sure plike dir exists
    assert os.path.exists( plike_dir )
    plike_dir = os.path.abspath(plike_dir) + r'/'

    cwd = os.getcwd()
    os.chdir(plike_dir)

    #Delete old plike.out if it exists
    if os.path.exists( plike_dir + 'plike.out' ):
        os.remove( plike_dir + 'plike.out' )

    #Make sure dcf file exists
    assert os.path.exists(dcf_fname)
    dcf_fname = os.path.abspath(dcf_fname)


    #Make sure there are two lag bounds
    if len(lag_bounds) != 2:
        print('Lag bounds: ', lag_bounds)
        raise ValueError('Must provide two lag bounds.')


    #Make file with the arguments to pass to plike
    with open('args.txt', 'w') as f:
        f.write(dcf_fname + '\n')
        f.write(str(lag_bounds[0]) + '\n')
        f.write(str(lag_bounds[1]))


    if verbose:
        print('Executing PLIKE')

    exec_str = './plike < args.txt'
    res = subprocess.Popen(exec_str, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    output, error = res.communicate()
    if res.returncode != 0:
        raise Exception("File handling failed %d %s %s" % (res.returncode, output, error))


    #Delete args.txt
    os.remove('args.txt')


    os.chdir(cwd)

    return


def get_zdcf(input_dir, fname1, fname2, out_dir, prefix='zdcf', num_MC=500, minpts=0,
             uniform_sampling=False, autocf=False, omit_zero_lags=True,
             sparse='auto', sep=',', verbose=False):

    """
    Runs the pyZDCF algorithm to compute the Z-Transformed Discrete Correlation Function (ZDCF).
    For more information on the algorithm, see pyZDCF (https://pyzdcf.readthedocs.io/).
    The algorithm will take in two light curves and output a file containing the ZDCF ('zdcf.dcf')
    in the specified output directory.


    Parameters
    ----------

    input_dir : str
            Path to the directory containing the light curves.

    fname1 : str
            Name of the first light curve file.

    fname2 : str
            Name of the second light curve file.

    out_dir : str
            Path to the directory to place the output ZDCF file.

    num_MC : float, optional
            The number of Monte Carlo simulations to run. Default is 500.

    minpts : int, optional
            The minimum number of points to use in each bin when computing the ZDCF.
            Must be larger than 11. If set to 0, it will be set to 11. Default is 0.

    uniform_sampling: bool, optional
            If True, the light curves will be assumed to be uniformly sampled.
            Default is ``False``.

    autocf : bool, optional
            If True, the auto-correlation function for the first light curve will be computed.
            If False, the ZDCF will be computed between the light curves.
            Default is ``False``.

    omit_zero_lags : bool, optional
            If True, will omit the points with zero lags when computing the ZDCF.
            Default is ``True``.

    sparse : (bool, str), optional
            Determines whether to use a sparse matrix implementation for reduced RAM usage.
            This feature is suitable for longer light curves (> 3000 data points). If True, will
            use sparse matrix implementation. If set to 'auto', will use sparse matrix implementation
            if there are more than 3000 data points per light curve. Default is 'auto'.

    sep : str, optional
            The delimiter used in the light curve files. Default is ',' for CSV files.

    verbose : bool, optional
            If ``True``, will output progress of pyZDCF. Default is ``False``.



    Returns
    -------

    dcf_df : pandas.DataFrame
        A pandas DataFrame object containing the ZDCF and its errors. The
        columns within the DataFrame match the order of columns within the
        output ZDCF file 'zdcf.dcf'.

    """


    #Files need to be csv
    params = dict(
        autocf = autocf,
        prefix = prefix,
        uniform_sampling = uniform_sampling,
        omit_zero_lags = True,
        minpts = 0,
        num_MC = num_MC,
        lc1_name = fname1,
        lc2_name = fname2
    )

    dcf_df = pyzdcf(input_dir=input_dir, output_dir=out_dir, intr=False,
                    verbose=verbose, parameters=params, sep=sep, sparse=sparse )

    return dcf_df



import multiprocessing as mp
from functools import partial

import numpy as np
import scipy.stats as sst


#For multiprocessing
def mp_map(func, arg, threads):

    if threads > 1:
        pool = mp.get_context('fork').Pool(threads)
        res = pool.map(func, arg)

        pool.close()
        pool.join()

    else:
        res = list(map(func, arg))

    return res

############################################################################
########################### Original pyCCF Code ############################
############################################################################

def corsig(r, v):
    '''
    Calculate the p value that a random uncorrelated sample can yield a
    correlation coefficient as large as or larger than the observed absolute
    value of r, where r is the correlation coefficient of the data (using
    t test, valid if v>=4)
    Ref1: http://janda.org/c10/Lectures/topic06/L24-significanceR.htm
    Ref2: http://vassarstats.net/textbook/ch4apx.html

    Inputs:
        r -- the correlation coefficient of the data
        v -- degree of freedom when calculating r: N-2 (hence N>2!!!)
    Outputs:
        pvalue
    '''
    r = float(r)
    v = float(v)

    r2 = r*r
    tst = r*np.sqrt(v/(1-r2))
    pvalue = sst.t.sf(tst, v) # sf: survival function -- 1-CDF
    return pvalue




def xcor(t1, y1, t2, y2, tlagmin, tlagmax, tunit, imode=0):
    '''
    Calculate cross-correlation function for unevenly
    sampling data.

    Inputs:
        t1 -- time for light curve 1, assume increase;
        y1 -- flux for light curve 1;
        t2 -- time for light curve 2, assume increase;
        y2 -- flux for light curve 2;
        tlagmin -- minimum time lag;
        tlagmax -- maximum time lag;
        tunit -- tau step;
        imode -- cross-correlation mode: 0, twice (default);
                 1, interpolate light curve 1;
                 2, interpolate light curve 2.

    Outputs:
        ccf -- correlation coefficient;
        tlag -- time lag (t2 - t1); positive values mean second
                  light curve lags the first light curve, as per convention.
                 (edit by kate, march 2016)
        npts -- number of data points used;
    '''
    if np.sum(np.diff(t1)<0.0)>0 or np.sum(np.diff(t2)<0.0)>0:
        raise Exception("The time of light curve 1 or light curve 2 is NOT INCREASING!!! Please check your data!!!")
    n1 = len(y1)
    n2 = len(y2)
    if n1<2 or n2<2:
        raise Exception("The light curve should contain at least 2 data points!!!")
    safe = tunit*0.1
    taulist12 = []
    taulist21 = []
    npts12 = []
    npts21 = []
    ccf12 = []  # interpolate 2
    ccf21 = []  # interpolate 1
    tau_max = tlagmax+safe
    # first interpolate 2
    if imode != 1:
        tau = tlagmin + 0.0 # if imode=1, skip the interpolate 2 step
    else:
        tau = tau_max + 0.0
    while tau < tau_max:
        t2new = t1 + tau
        selin = np.where((t2new>=np.min(t2))&(t2new<=np.max(t2)), True, False)
        knot = np.sum(selin)  # number of datapoints used
        if knot>0:
            y2new = np.interp(t2new[selin], t2, y2)

            y1sum = np.sum(y1[selin])
            y1sqsum = np.sum(y1[selin]*y1[selin])
            y2sum = np.sum(y2new)
            y2sqsum = np.sum(y2new*y2new)
            y1y2sum = np.sum(y1[selin]*y2new)

            fn = float(knot)
            rd1_sq = fn*y2sqsum - y2sum*y2sum
            rd2_sq = fn*y1sqsum - y1sum*y1sum
            if rd1_sq>0.0:
                rd1 = np.sqrt(rd1_sq)
            else:
                rd1 = 0.0
            if rd2_sq>0.0:
                rd2 = np.sqrt(rd2_sq)
            else:
                rd2 = 0.0

            if rd1*rd2==0.0:
                r = 0.0
            else:
                r = (fn*y1y2sum - y2sum*y1sum)/(rd1*rd2)
            ccf12.append(r)
            taulist12.append(tau)
            npts12.append(knot)
        tau += tunit
    # now interpolate 1
    if imode != 2:
        tau = tlagmin + 0.0
    else:
        tau = tau_max + 0.0
    while tau < tau_max:
        t1new = t2 - tau
        selin = np.where((t1new>=np.min(t1))&(t1new<=np.max(t1)), True, False)
        knot = np.sum(selin)  # number of datapoints used
        if knot>0:
            y1new = np.interp(t1new[selin], t1, y1)

            y2sum = np.sum(y2[selin])
            y2sqsum = np.sum(y2[selin]*y2[selin])
            y1sum = np.sum(y1new)
            y1sqsum = np.sum(y1new*y1new)
            y1y2sum = np.sum(y1new*y2[selin])

            fn = float(knot)
            rd1_sq = fn*y2sqsum - y2sum*y2sum
            rd2_sq = fn*y1sqsum - y1sum*y1sum
            if rd1_sq>0.0:
                rd1 = np.sqrt(rd1_sq)
            else:
                rd1 = 0.0
            if rd2_sq>0.0:
                rd2 = np.sqrt(rd2_sq)
            else:
                rd2 = 0.0

            if rd1*rd2==0.0:
                r = 0.0
            else:
                r = (fn*y1y2sum - y2sum*y1sum)/(rd1*rd2)
            ccf21.append(r)
            taulist21.append(tau)
            npts21.append(knot)
        tau += tunit

    # return results according to imode
    taulist12 = np.asarray(taulist12)
    npts12 = np.asarray(npts12)
    taulist21 = np.asarray(taulist21)
    npts21 = np.asarray(npts21)
    ccf12 = np.asarray(ccf12)
    ccf21 = np.asarray(ccf21)
    if imode==0:
        # make sure taulist12 and taulist21 have the same size!!!
        if np.array_equal(taulist12, taulist21):
            ccf = (ccf12 + ccf21)*0.5
            taulist = taulist12 + 0.0
            npts = npts12 + 0.0
        else:
            taulist = np.intersect1d(taulist12, taulist21)
            sel_cb12 = np.in1d(taulist12, taulist)
            sel_cb21 = np.in1d(taulist21, taulist)
            ccf = (ccf12[sel_cb12] + ccf21[sel_cb21])*0.5
            npts = (npts12[sel_cb12] + npts21[sel_cb21])*0.5
    elif imode==1:
        ccf = ccf21 + 0.0
        taulist = taulist21 + 0.0
        npts = npts21 + 0.0
    else:
        ccf = ccf12 + 0.0
        taulist = taulist12 + 0.0
        npts = npts12 + 0.0

    return ccf, taulist, npts




def peakcent(t1, y1, t2, y2, tlagmin, tlagmax, tunit, thres=0.8, siglevel=0.95, imode=0, sigmode = 0.2):
    '''
    Calculate peak time lag and centroid based on the cross-correlation
    function for unevenly sampling data.

    Inputs:
        t1 -- time for light curve 1, assume increase;
        y1 -- flux for light curve 1;
        t2 -- time for light curve 2, assume increase;
        y2 -- flux for light curve 2;
        tlagmin -- minimum time lag;
        tlagmax -- maximum time lag;
        tunit -- tau step;
        thres -- lower limit of correlation coefficient when
                 calculate centroid, default is 0.8;
        siglevel -- the required significant level of the
                 correlation coefficient;
        imode -- cross-correlation mode: 0, twice (default);
                 1, interpolate light curve 1;
                 2, interpolate light curve 2.
        sigmode -- how to deal with significance:
                Will use r = input value as the minimum correlation coefficient to consider (default = 0.2).
                0: Will use a p-test to assign significance to peak and discard peaks that are below
                the significance threshold (depends on number of points included and r).

    Outputs:
        tlag_peak -- time lag based on the peak argument;
        status_peak -- peak status (1, constrained; 0, unconstrained);
        tlag_centroid -- time lag for centroid;
        status_centroid -- centroid status (1, constrained; 0, unconstrained);
    '''
    alpha = 1.0 - siglevel  # probability threshold to reject: no correlation hypothesis

    ccf_pack = xcor(t1, y1, t2, y2, tlagmin, tlagmax, tunit, imode)
    max_indx = np.argmax(ccf_pack[0])
    max_rval = ccf_pack[0][max_indx]
    if ccf_pack[2][max_indx]>2.0:
        peak_pvalue = corsig(ccf_pack[0][max_indx], float(ccf_pack[2][max_indx]-2.0))
    else:
        peak_pvalue = 1.0 # significance level
    # ccf peaks --- excluding all with r < 0.2 instead of using p-value test.
    if sigmode > 0:
        #print 'Using minimum r coefficient instead of significance test.'
        #Check and see if the max r is on the edge of the CCF. Fail it if so.
        if max_rval >= sigmode and ccf_pack[1][max_indx] > tlagmin and ccf_pack[1][max_indx] < tlagmax:
            tlag_peak = ccf_pack[1][max_indx]
            max_rval = max_rval
            status_peak = 1
            status_rval = 1
            status_centroid = 0 # if lag is well determined, we will change status_centroid to 1
            tlag_centroid = -9999.0
        else:
            max_rval = -9999.0
            tlag_peak = -9999.0
            tlag_centroid = -9999.0
            status_peak = 0
            status_rval = 0
            status_centroid = 0
    else:
        # ccf peaks-- Eric's method using a p-value test (usually not using)
        #Check and see if the max r is on the edge of the CCF. Fail it if so.
        if peak_pvalue<alpha and ccf_pack[1][max_indx] > tlagmin and ccf_pack[1][max_indx] < tlagmax:
            tlag_peak = ccf_pack[1][max_indx]
            max_rval = max_rval
            status_peak = 1
            status_rval = 1
            status_centroid = 0 # if lag is well determined, we will change status_centroid to 1
            tlag_centroid = -9999.0
        else:
            max_rval = -9999.0
            tlag_peak = -9999.0
            tlag_centroid = -9999.0
            status_peak = 0
            status_rval = 0
            status_centroid = 0
    #If the peak succeeds, calculate centroid:
    if status_peak == 1:
        rcent = thres*max_rval
        # find out the range of centroid around the primary peak
        rdif_neg = np.where(ccf_pack[0]-rcent<0.0, True, False)
        tlag_rneg = ccf_pack[1][rdif_neg] - tlag_peak
        tlag_leftall = np.abs(tlag_rneg[tlag_rneg<0.0])
        tlag_rightall = np.abs(tlag_rneg[tlag_rneg>0.0])
        if len(tlag_leftall)>0 and len(tlag_rightall)>0:
            tlag_left = tlag_peak - np.min(tlag_leftall) # the left edge of the centroid around the primary peak
            tlag_right = tlag_peak + np.min(tlag_rightall) # the right edge of the centroid around the primary peak
            if tlag_left>=np.min(ccf_pack[1]) and tlag_right<=np.max(ccf_pack[1]):
                # centroids
                selcen = np.where((ccf_pack[1]>tlag_left)&(ccf_pack[1]<tlag_right),True,False)
                if np.sum(selcen)>0:
                    tlag_centroid = np.sum(ccf_pack[0][selcen]*ccf_pack[1][selcen])/np.sum(ccf_pack[0][selcen])
                    status_centroid = 1
    # end of centroid calculation
    #Now, if the centroid fails, re-set the peak status to 0 because we don't want to report a peak without a centroid!
    if status_centroid == 0:
        status_peak = 0
        tlag_peak = -9999.0
        max_rval = -9999.0
        status_rval = 0
    #print tlag_peak, status_peak, tlag_centroid, status_centroid, max_rval, status_rval
    return tlag_peak, status_peak, tlag_centroid, status_centroid, ccf_pack, max_rval, status_rval, peak_pvalue




def xcor_mc_loop(mcmode, t1, y1, dy1, t2, y2, dy2, tlagmin, tlagmax, tunit,
                 numt1, numt2,
                 thres=0.8, siglevel=0.95,
                 imode=0, sigmode=0.2):

    if mcmode!=2:
        # RSS resample light curve 1
        mycheck = True #make sure len(t1_rss)>1
        while mycheck:
            indx1 = np.random.randint(0, numt1, numt1)
            unique1, counts1 = np.unique(indx1, return_counts=True) # sorted unique value
            t1_rss = t1[unique1]
            y1_rss = y1[unique1]
            dy1_rss = dy1[unique1]/np.sqrt(counts1)
            if len(t1_rss)>1:
                # keep running unless len(t1_rss)>1
                mycheck = False
        # RSS resample light curve 2
        mycheck = True #make sure len(t2_rss)>1
        while mycheck:
            indx2 = np.random.randint(0, numt2, numt2)
            unique2, counts2 = np.unique(indx2, return_counts=True)
            t2_rss = t2[unique2]
            y2_rss = y2[unique2]
            dy2_rss = dy2[unique2]/np.sqrt(counts2)
            if len(t2_rss)>1:
                # keep running unless len(t2_rss)>1
                mycheck = False

    else:
        # do not apply RSS resample, rss light curve equals to original one
        t1_rss = t1 + 0.0
        y1_rss = y1 + 0.0
        dy1_rss = dy1 + 0.0

        t2_rss = t2 + 0.0
        y2_rss = y2 + 0.0
        dy2_rss = dy2 + 0.0


    if mcmode!=1:
        # measurement error perturbation
        t1_fr = t1_rss + 0.0
        y1_fr = np.random.normal(y1_rss, dy1_rss)
        t2_fr = t2_rss + 0.0
        y2_fr = np.random.normal(y2_rss, dy2_rss)
    else:
        # do not aplly the error perturbation
        t1_fr = t1_rss + 0.0
        y1_fr = y1_rss + 0.0
        t2_fr = t2_rss + 0.0
        y2_fr = y2_rss + 0.0

    # perform CCF
    pc_pack = peakcent(t1_fr, y1_fr, t2_fr, y2_fr, tlagmin, tlagmax, tunit, thres, siglevel, imode, sigmode = sigmode)

    # ccf peaks
    if pc_pack[1] == 1:
        tau_peak = pc_pack[0]
        pval = pc_pack[7]
        success_peak = True
    else:
        tau_peak = np.nan
        pval = np.nan
        success_peak = False


    # ccf centroids
    if pc_pack[3] == 1:
        tau_centroid = pc_pack[2]
        success_centroid = True
    else:
        tau_centroid = np.nan
        success_centroid = False


    # max_rvalues
    if pc_pack[6] == 1:
        max_rval = pc_pack[5]
        success_rval = True
    else:
        max_rval = np.nan
        success_rval = False

    return tau_peak, pval, success_peak, tau_centroid, success_centroid, max_rval, success_rval




def xcor_mc(t1, y1, dy1, t2, y2, dy2, tlagmin, tlagmax, tunit,
            thres=0.8, siglevel=0.95, imode=0, nsim=2048, mcmode=0,
            sigmode=0.2, threads=1, verbose=True):
    '''
    Calculate the uncertainty for the cross-correlation peak.

    Inputs:
        t1 -- time for light curve 1, assume increase;
        y1 -- flux for light curve 1;
        dy1 -- flux uncertainty for light curve 1;
        t2 -- time for light curve 2, assume increase;
        y2 -- flux for light curve 2;
        dy2 -- flux uncertainty for light curve 2;
        tlagmin -- minimum time lag;
        tlagmax -- maximum time lag;
        tunit -- tau step;
        thres -- lower limit of correlation coefficient when
                 calculate centroid, default is 0.8;
        siglevel -- the required significant level of the
                 correlation coefficient;
        imode -- cross-correlation mode: 0, twice (default);
                 1, interpolate light curve 1;
                 2, interpolate light curve 2.
        nsim -- MC simulation trials;
        mcmode -- MC mode: 0, RSS plus FR
                  1, RSS only
                  2, FR only
        sigmode -- How to exclude non-significant peaks:
                  Will exclude all peaks with r < input value
                  0 will exclude all peaks based on p-value significance test.

    Outputs:
        tlags_peak -- tlag of peak distribution;
        tlags_centroid -- tlag of centroid distribution;
        nsuccess_peak -- success times in finding peaks;
        nfail_peak -- fail times in finding peaks;
        nsuccess_centroid -- success times in calculating centroid;
        nfail_centroid -- fail times in calculating centroid.
    '''
    if np.sum(np.diff(t1)<0.0)>0 or np.sum(np.diff(t2)<0.0)>0:
        raise Exception("The time of light curve 1 or light curve 2 is NOT INCREASING!!! Please check your data!!!")
    numt1 = len(t1)
    numt2 = len(t2)
    if numt1<2 or numt2<2:
        raise Exception("The light curve should contain at least 2 data points!!!")

    pyccf_func = partial( xcor_mc_loop,
                         t1=t1, y1=y1, dy1=dy1, t2=t2, y2=y2, dy2=dy2,
                         tlagmin=tlagmin, tlagmax=tlagmax, tunit=tunit,
                         numt1=numt1, numt2=numt2,
                         thres=thres, siglevel=siglevel,
                         imode=imode, sigmode=sigmode)


    arg = np.full(nsim, mcmode)
    res = mp_map( pyccf_func, arg, threads )

    tlags_peak, pvals, success_peak, tlags_centroid, success_centroid, max_rvals, success_rval = np.array(res).T

    nsuccess_peak = len( np.argwhere( success_peak ).T[0] )
    nfail_peak = len( np.argwhere( success_peak == False ).T[0] )

    nsuccess_centroid = len( np.argwhere( success_centroid ).T[0] )
    nfail_centroid = len( np.argwhere( success_centroid == False ).T[0] )

    nsuccess_rvals = len( np.argwhere( success_rval ).T[0] )
    nfail_rvals = len( np.argwhere( success_rval == False ).T[0] )



    tlags_peak = np.asarray(tlags_peak)
    tlags_centroid = np.asarray(tlags_centroid)

    #Get rid of NaNs
    if np.all( np.isnan(tlags_peak) ):
        tlags_peak = np.array([np.nan])
    else:
        tlags_peak = tlags_peak[~np.isnan(tlags_peak)]

    if np.all( np.isnan(tlags_centroid) ):
        tlags_centroid = np.array([np.nan])
    else:
        tlags_centroid = tlags_centroid[~np.isnan(tlags_centroid)]



    if verbose:
        print( 'Failed centroids: ', nfail_centroid )
        print( 'Failed peaks: ', nfail_peak )

    return tlags_peak, tlags_centroid, nsuccess_peak, nfail_peak, nsuccess_centroid, nfail_centroid, max_rvals, nfail_rvals, pvals


############################################################################
############################################################################
############################################################################

def get_pyccf_lags(fname1, fname2,
                   lag_bounds=None, interp=None,
                   nsim=1000, mcmode=0, sigmode=.2, thres=.8,
                   threads=1, verbose=False):

    """Obtain time lags from the correlation between two light curves using the pyCCF algorithm: https://ui.adsabs.harvard.edu/abs/2018ascl.soft05032S/abstract.
    In short, pyCCF creates the Interpolated Cross-Correlation Function (ICCF) between two light curves,
    and then samples the peak and centroid of the ICCF using Monte Carlo simulations with
    Flux Randomization (FR) and/or Random Subset Resampling (RSS). These Monte Carlo simulations are then
    used to create the cross-correlation centroid distribution (CCCD) and peak distribution (CCPD).



    Parameters
    ----------

    fname1 : str
        Path to the first light curve file .

    fname2 : str
        Path to the second light curve file .

    file_fmt : str, optional
        Format of the file. Default is 'csv'.

    lag_bounds : (2,) array_like, optional
        The bounds of times to search for the lag. The first element is the minimum lag and the second is the maximum.
        If set to ``None", the lag bounds will be set to (-baseline, baseline). The default is ``None``.

    interp : float, optional
        The interval with which pyCCF will interpolate the ligh curves to form the ICCF. This value must be
        shorter than the average cadence of the ligh curves. Setting this value too low can introduce noise.
        If set to ``None``, will be set to half of the average cadence of the light curves. The default is ``None``.

    nsim : int, optional
        The number of Monte Carlo simulations to run. The default is 1000.

    mcmode : int, optional
        The type of resampling to do for the Monte Carlo Simulations. 0 performs both FR and RSS, 1 performs FR, and 2 performs RSS.
        The default is 0.

    sigmode : float, optional
        The threshold for considering a measurement in the ICCF significant when computing peaks and centroids. Must be within the
        interval (0,1). All peaks and centroids with correlation coefficient r_max <= sigmode will be considered as "failed".
        If set to 0, will exclude all peaks based on a p-value significance test (see pyCCF documentation). The default is 0.2.

    thres : float, optional
        The lower limit of correlation coefficient used when calculating the centroid of the ICCF. Must be within
        the interval (0,1). The default is 0.8.


    .. note:: Both light curve files must be in CSV format with the following columns in order: time, value, uncertainty.



    Returns
    -------

    result: dict
        Dict of output results, containing:

        * 'CCF' : list of floats
            The ICCF.

        * 'CCF_lags' : list of floats
            The lags corresponding to the ICCF.

        * 'centroid' : float
            The median of the CCCD.

        * 'centroid_err_hi' : float
            The upper error of the centroid.

        * 'centroid_err_lo' : float
            The lower error of the centroid.

        * 'peak' : float
            The median of the CCPD.

        * 'peak_err_hi' : float
            The upper error of the peak.

        * 'peak_err_lo' : float
            The lower error of the peak.

        * 'CCCD_lags' : list of floats
            The CCCD.

        * 'CCPD_lags' : list of floats
            The CCPD.

    """

    #Files need to be csv
    x1, y1, yerr1 = np.loadtxt( fname1, delimiter=',', unpack=True, usecols=[0,1,2] )
    x2, y2, yerr2 = np.loadtxt( fname2, delimiter=',', unpack=True, usecols=[0,1,2] )

    #Sort by time
    sort_ind = np.argsort(x1)
    x1 = x1[sort_ind]
    y1 = y1[sort_ind]
    yerr1 = yerr1[sort_ind]

    sort_ind = np.argsort(x2)
    x2 = x2[sort_ind]
    y2 = y2[sort_ind]
    yerr2 = yerr2[sort_ind]

    #Get mean cadence and baseline
    diff1 = np.diff(x1)
    diff2 = np.diff(x2)
    mean_diff = np.mean( [ np.mean(diff1), np.mean(diff2) ] )
    baseline = np.max([ x1[-1], x2[-1] ]) - np.min([ x1[0], x2[0] ])

    #Define pyCCF parameters
    if lag_bounds is None:
        lag_bounds = [-baseline, baseline]

    if interp is None:
       interp = mean_diff/2


    if np.any( np.abs(lag_bounds) > baseline ):
        print('Lag bounds are larger than baseline')


    #Run algorithm
    _, status_peak, _, status_centroid, ccf_pack, \
        _, status_rval, _ = peakcent(x1, y1, x2, y2, lag_bounds[0],
                                           lag_bounds[1], interp)

    # assert status_peak == 1
    # assert status_centroid == 1

    tlags_peak, tlags_centroid, nsuccess_peak, nfail_peak, \
        nsuccess_centroid, nfail_centroid, max_rvals, nfail_rvals, \
            pvals = xcor_mc(x1, y1, np.abs(yerr1), x2, y2,
                                np.abs(yerr2), lag_bounds[0], lag_bounds[1], interp,
                                nsim=nsim, mcmode=mcmode, sigmode=sigmode,
                                thres=thres, threads=threads, verbose=verbose)

    ccf = ccf_pack[0]
    ccf_lags = ccf_pack[1]

    #Get peak and centroid values/uncertainties
    cent_med = np.median( tlags_centroid )
    cent_hi = np.percentile( tlags_centroid, 84 )
    cent_lo = np.percentile( tlags_centroid, 16 )

    peak_med = np.median( tlags_peak )
    peak_hi = np.percentile( tlags_peak, 84 )
    peak_lo = np.percentile( tlags_peak, 16 )


    return {
        'CCF': ccf,
        'CCF_lags': ccf_lags,
        'centroid': cent_med,
        'centroid_err_hi': cent_hi - cent_med,
        'centroid_err_lo': cent_med - cent_lo,
        'peak': peak_med,
        'peak_err_hi': peak_hi - peak_med,
        'peak_err_lo': peak_med - peak_lo,
        'CCCD_lags': tlags_centroid,
        'CCPD_lags': tlags_peak
    }



import itertools
import multiprocessing as mp
import os
import shutil
from functools import partial

import numpy as np
from astropy.table import Table

import pypetal.pyzdcf.utils as utils
from pypetal.pyzdcf.plotting import plot_pyzdcf_results
from pypetal.utils import defaults


#For multiprocessing
def mp_map(func, args, threads):
    n_inputs = len(args)

    if (threads > 1) & (n_inputs > 1):
        pool = mp.get_context('fork').Pool(threads)
        res = pool.starmap(func, args)

        pool.close()
        pool.join()

    else:
        res = list(itertools.starmap(func, args))

    return res



def pyzdcf_tot(cont_fname, line_fnames, line_names, output_dir,
               general_kwargs, kwargs):


    if line_fnames is str:
        line_fnames = [line_fnames]
        line_names = [line_names]

    #--------------------------------------------------
    #Read general kwargs

    verbose = general_kwargs['verbose']
    plot = general_kwargs['plot']
    time_unit = general_kwargs['time_unit']
    lc_unit = general_kwargs['lc_unit']
    lag_bounds = general_kwargs['lag_bounds']
    threads = general_kwargs['threads']

    #--------------------------------------------------
    #Read kwargs

    nsim, minpts, uniform_sampling, omit_zero_lags, \
        sparse, prefix, run_plike, plike_dir = defaults.set_pyzdcf(kwargs,
                                                                   np.hstack([ [cont_fname], line_fnames ])
                                                                   )

    #-------------------------------------------
    if (run_plike) & (plike_dir is None):
        print('Error: plike_dir must be specified if run_plike=True')
        print('Skipping PLIKE')
        run_plike = False

    input_dir = os.path.dirname( os.path.realpath(cont_fname) )

    for i in range(len(line_fnames)):
        if input_dir != os.path.dirname( os.path.realpath(line_fnames[i]) ):
            print('ERROR: All light curve files must be in the same directory')
            return {}


    if verbose:

        txt_str = """
Running pyZDCF
----------------------
nsim: {}
minpts: {}
uniform_sampling: {}
omit_zero_lags: {}
sparse: {}
prefix: {}
run_plike: {}
plike_dir: {}
----------------------
        """.format( nsim, minpts, uniform_sampling, omit_zero_lags,
                    sparse, prefix, run_plike, plike_dir)

        print(txt_str)



    input_dir += r'/'
    cont_fname_short = os.path.basename(cont_fname)
    line_fnames_short = [ os.path.basename(x) for x in line_fnames ]


    pyzdcf_func = partial( utils.get_zdcf, num_MC=nsim, minpts=minpts,
                           uniform_sampling=uniform_sampling, omit_zero_lags=omit_zero_lags,
                           sparse=sparse, sep=',', verbose=False)

    arg1 = np.full( len(line_fnames), input_dir )
    arg2 = np.full( len(line_fnames), cont_fname_short )
    arg3 = line_fnames_short
    arg4 = [ output_dir + x + r'/pyzdcf/' for x in line_names[1:] ]
    arg5 = [ x + '_' + prefix for x in line_names[1:] ]

    args = list(zip(arg1, arg2, arg3, arg4, arg5))
    res_tot = mp_map(pyzdcf_func, args, threads)

    plike_tot = []
    for i in range(len(line_fnames)):

        x1, y1, yerr1 = np.loadtxt( cont_fname, delimiter=',', unpack=True, usecols=[0,1,2] )
        x2, y2, yerr2 = np.loadtxt( line_fnames[i], delimiter=',', unpack=True, usecols=[0,1,2] )

        plike_dict = None
        if run_plike:

            utils.run_plike( arg4[i] + arg5[i] + '.dcf', lag_bounds[i], plike_dir,
                            verbose=verbose)

            plike_fname = plike_dir + 'plike.out'
            output_fname = output_dir + line_names[i+1] + r'/pyzdcf/' + line_names[i+1] + '_plike.out'
            assert os.path.exists( os.path.dirname(output_fname) )

            shutil.move( plike_fname, output_fname )
            plike_dat = Table.read( output_fname,
                                    format='ascii',
                                    names=['num', 'lag', 'r', '-dr', '+dr', 'likelihood'])


            file = open(output_fname, 'r')
            output_str = list(file)[-3:]

            ml_lag = float( output_str[1].split()[7] )
            ml_lag_err_hi = np.abs( float( output_str[1].split()[8] )  )
            ml_lag_err_lo = np.abs( float( output_str[1].split()[9] )  )

            plike_dict = {
                'output': plike_dat,
                'ML_lag': ml_lag,
                'ML_lag_err_hi': ml_lag_err_hi,
                'ML_lag_err_lo': ml_lag_err_lo
            }

            plike_tot.append( plike_dict )

        plot_pyzdcf_results(x1, y1, yerr1, x2, y2, yerr2,
                                     res_tot[i], plike_dict,
                                     time_unit=time_unit, lc_unit=[lc_unit[0], lc_unit[i+1]],
                                     lc_names=[line_names[0], line_names[i+1]],
                                     fname=output_dir + line_names[i+1] + r'/pyzdcf/' + line_names[i+1] + '_zdcf.pdf',
                                     show=plot)

    return res_tot, plike_tot



import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from linmix import LinMix

from pypetal.utils.defaults import set_detrend
from pypetal.utils.petalio import write_data

mpl.rcParams['xtick.minor.visible'] = True
mpl.rcParams['xtick.top'] = True
mpl.rcParams['xtick.direction'] = 'in'

mpl.rcParams['ytick.minor.visible'] = True
mpl.rcParams['ytick.right'] = True
mpl.rcParams['ytick.direction'] = 'in'

mpl.rcParams["figure.autolayout"] = False

mpl.rcParams['savefig.dpi'] = 300
mpl.rcParams['savefig.format'] = 'pdf'



def detrend(x, y, yerr, K=2, parallelize=False,
            nchain=4, miniter=5000, maxiter=10000,
            time_unit='d', lc_unit='',
            output_dir=None, verbose=False, plot=False):


    """Detrend a light curve using a linear model from LinMix.

    Parameters
    ----------

    x : list of float
        Time values of the light curve.

    y : list of float
        Values of the light curve.

    yerr : list of float
        Uncertainties in the light curve.

    K : int, optional
        The number of Gaussians to use in the LinMix model. Default is 2.

    parallelize : bool, optional
        Whether to parallelize the MCMC chains. Default is ``False``.

    nchain : int, optional
        The number of MCMC chains to run. Default is 4.

    miniter : int, optional
        The minimum number of iterations to run. Default is 5000.

    maxiter : int, optional
        The maximum number of iterations to run. Default is 10000.

    time_unit : str, optional
        The unit of time for the light curve. Default is 'd'.

    lc_unit : str, optional
        The unit of the light curve. Default is ''.

    output_dir : str, optional
        The directory to save the detrending figure to. Default is ``None``.

    verbose : bool, optional
        Whether to print the results of the detrending. Default is ``False``.

    plot : bool, optional
        Whether to plot the detrending figure. Default is ``False``.


    Returns
    -------

    y_dt : list of float
        The detrended light curve.

    m_res: list of float
        The fitted value and uncertainty in the slope of the linear model, given as [lower error, value, upper error].

    b_res: list of float
        The fitted value and uncertainty in the intercept of the linear model, given as [lower error, value, upper error].

    sigsqr_res: list of float
        The fitted value and uncertainty in the variance of the linear model, given as [lower error, value, upper error].

    """



    if lc_unit == '':
        flux_txt = 'Flux'
    elif lc_unit == 'mag':
        flux_txt = 'Magnitude'
    else:
        flux_txt = 'Flux [{}]'.format(lc_unit)


    lm = LinMix(x, y, ysig=yerr, K=K, nchains=nchain, parallelize=parallelize)
    lm.run_mcmc(miniter=miniter, maxiter=maxiter, silent=True)

    b_chains = []
    m_chains = []
    sigsqr_chains = []

    for i in range(len(lm.chain)):
        b_chains.append( lm.chain[i][0] )
        m_chains.append( lm.chain[i][1] )
        sigsqr_chains.append( lm.chain[i][1] )


    b_med = np.median(b_chains)
    b_err_hi = np.percentile(b_chains, 84) - b_med
    b_err_lo = b_med - np.percentile(b_chains, 16)

    m_med = np.median(m_chains)
    m_err_hi = np.percentile(m_chains, 84) - m_med
    m_err_lo = b_med - np.percentile(m_chains, 16)

    sigsqr_med = np.median(sigsqr_chains)
    sigsqr_err_hi = np.percentile(sigsqr_chains, 84) - sigsqr_med
    sigsqr_err_lo = sigsqr_med - np.percentile(sigsqr_chains, 16)


    if verbose:
        print('m = {:.3f} + {:.3f} - {:.3f}'.format(m_med, m_err_hi, m_err_lo) )
        print('b = {:.3f} + {:.3f} - {:.3f}'.format(b_med, b_err_hi, b_err_lo) )
        print('sigsqr = {:.3f} + {:.3f} - {:.3f}'.format(sigsqr_med, sigsqr_err_hi, sigsqr_err_lo) )



    y0_vals = []
    for i in range(len(lm.chain)):
        m_val = lm.chain[i][1]
        b_val = lm.chain[i][0]
        x0 = np.min(x)

        y0_vals.append( m_val*x0 + b_val )


    lo_ind = np.argmin(y0_vals)
    m_lo = lm.chain[lo_ind][1]
    b_lo = lm.chain[lo_ind][0]

    hi_ind = np.argmax(y0_vals)
    m_hi = lm.chain[hi_ind][1]
    b_hi = lm.chain[hi_ind][0]



    xsim = np.linspace(x.min(), x.max(), 100)
    ysim_avg = m_med*xsim + b_med
    ysim_lo = m_lo*xsim + b_lo
    ysim_hi = m_hi*xsim + b_hi



    fig, ax = plt.subplots()

    _, _, bars = ax.errorbar(x, y, yerr, fmt='.k')
    [bar.set_alpha(.3) for bar in bars]


    ax.plot(xsim, ysim_avg, 'r')
    ax.fill_between(xsim, ysim_lo, ysim_hi, color='r', alpha=.3)

    if lc_unit == 'mag':
        ax.invert_yaxis()

    ax.set_ylabel(flux_txt)
    ax.set_xlabel('Time [{}]'.format(time_unit))

    ax.tick_params('both', which='major', length=8)
    ax.tick_params('both', which='minor', length=3)

    if output_dir is not None:
        plt.savefig(output_dir + 'detrend.pdf', dpi=200, bbox_inches='tight')

    if plot:
        plt.show()

    plt.cla()
    plt.clf()
    plt.close()


    y_dt = y - (m_med*x + b_med)

    return y_dt, [m_err_lo, m_med, m_err_hi], \
        [b_err_lo, b_med, b_err_hi], \
        [sigsqr_err_lo, sigsqr_med, sigsqr_err_hi]









def detrend_tot(output_dir, cont_fname, line_fnames, line_names, general_kwargs, kwargs):

    verbose = general_kwargs['verbose']
    plot = general_kwargs['plot']
    threads = general_kwargs['threads']
    time_unit = general_kwargs['time_unit']
    lc_unit = general_kwargs['lc_unit']


    K, nchain, miniter, maxiter = set_detrend(kwargs)


    fnames_tot = np.hstack([ [cont_fname], line_fnames])

    if threads > 1:
        parallelize = True
    else:
        parallelize = False


    txt = """
Running detrending
-------------------
parallelize: {}
K: {}
nchains: {}
miniter: {}
maxiter: {}
-------------------
    """.format(parallelize, K, nchain, miniter, maxiter)

    if verbose:
        print(txt)


    m_vals = []
    b_vals = []
    sigsqr_vals = []

    for i, fname in enumerate(fnames_tot):
        x, y, yerr = np.loadtxt( fname, unpack=True, usecols=[0,1,2], delimiter=',' )

        y_dt, m_res, b_res, sigsqr_res = detrend(x, y, yerr, K=K,
                                                parallelize=parallelize,
                                                time_unit=time_unit, lc_unit=lc_unit[i],
                                                output_dir=output_dir + line_names[i] + r'/',
                                                verbose=verbose, plot=plot)

        write_data( [x, y_dt, yerr], output_dir + 'processed_lcs/' + line_names[i] + '_data.dat' )

        m_vals.append(m_res)
        b_vals.append(b_res)
        sigsqr_vals.append(sigsqr_res)


    output = {
        'm': m_vals,
        'b': b_vals,
        'sigsqr': sigsqr_vals
    }

    return output



from .defaults import *
from .detrending import *
from .load import *
from .petalio import *

__all__ = ["pypetal.utils.defaults", "pypetal.utils.detrending", "pypetal.utils.load", "pypetal.utils.petalio"]



import glob
import os

import numpy as np
import toml


def str2none(x):
    if x == 'None':
        return None
    else:
        return x

#########################################################################################
################################# ASSIST FUNCTIONS ######################################
#########################################################################################

def make_toml(output_dir, fnames,
              run_arr, param_arr,
              line_names=None, filename=None):

    tot_keys = ['drw_rej', 'detrend', 'pyccf', 'pyzdcf', 'pyroa', 'javelin', 'weighting']
    toml_dict = {}


    #Inputs
    toml_dict['inputs'] = {}
    toml_dict['inputs']['output_dir'] = output_dir
    toml_dict['inputs']['filenames'] = fnames

    if line_names is not None:
        toml_dict['inputs']['line_names'] = line_names





    toml_dict['params'] = {}

    #General parameters
    if param_arr[0] != {}:
        toml_dict['params']['general'] = param_arr[0]


    #Module parameters
    for i in range(len(run_arr)):
        if run_arr[i]:
            toml_dict['params'][tot_keys[i]] = param_arr[i+1]
            toml_dict['params'][tot_keys[i]]['run'] = True

    #Write to file
    if filename is not None:
        toml.dump(toml_dict, open(filename, 'w+'))

    return toml_dict





def get_toml_modules(filename):

    tot_keys = ['drw_rej', 'detrend', 'pyccf', 'pyzdcf', 'pyroa', 'javelin', 'weighting']
    run_arr = []

    toml_dat = toml.load(filename)['params']

    for key in tot_keys:
        if key in toml_dat.keys():
            if 'run' in toml_dat[key]:
                run_val = toml_dat[key]['run']
            else:
                run_val = False
        else:
            run_val = False


        run_arr.append(run_val)

    return run_arr




def get_toml_params(filename, run_arr):

    tot_keys = ['drw_rej', 'detrend', 'pyccf', 'pyzdcf', 'pyroa', 'javelin', 'weighting']
    param_arr = []


    toml_dat = toml.load(filename)['params']

    #General parameters
    if 'general' in toml_dat.keys():
        param_arr.append(toml_dat['general'])
    else:
        param_arr.append({})


    #Module parameters
    for i in range(len(run_arr)):

        if run_arr[i]:
            param_arr.append(toml_dat[tot_keys[i]])
        else:
            param_arr.append({})

    return param_arr








def get_toml_inputs(filename):

    toml_dat = toml.load(filename)

    assert 'inputs' in toml_dat.keys(), 'No inputs found in toml file'
    assert 'output_dir' in toml_dat['inputs'].keys(), 'No output directory found in toml file'
    assert 'filenames' in toml_dat['inputs'].keys(), 'No filenames found in toml file'


    #################################
    #Output directory

    output_dir = os.path.abspath(toml_dat['inputs']['output_dir']) + r'/'


    #################################
    #Filenames

    #If input is a list
    if isinstance( toml_dat['inputs']['filenames'], list ):
        for i in range(len(toml_dat['inputs']['filenames'])):
            assert os.path.isfile(toml_dat['inputs']['filenames'][i]), 'File not found: ' + toml_dat['inputs']['filenames'][i]

        filenames = toml_dat['inputs']['filenames']



    #If input is a string
    if isinstance( toml_dat['inputs']['filenames'], str ):
        exists = os.path.exists(toml_dat['inputs']['filenames'])
        isdir = os.path.isdir(toml_dat['inputs']['filenames'])


        #Assume all files in directory are inputs
        if exists and isdir:
            filenames = glob.glob(toml_dat['inputs']['filenames']+'*')

        #Assume input is a glob pattern
        elif exists and (not isdir):
            filenames = glob.glob(toml_dat['inputs']['filenames'])
            assert len(filenames) > 0, 'No files found matching glob pattern: ' + toml_dat['inputs']['filenames']


        for i in range(len(filenames)):
            assert os.path.isfile(filenames[i]), 'File not found: ' + filenames[i]




    #################################
    #Line names

    if str2none(toml_dat['inputs']['line_names']) is None:
        line_names = []
        for i in range(len(filenames)):
            line_names.append('Line {}'.format(i+1))

    else:
        assert isinstance( toml_dat['inputs']['line_names'], list ), 'Line names must be a list'
        line_names = toml_dat['inputs']['line_names']

    return output_dir, filenames, line_names



#########################################################################################
################################ TO RUN SINGLE OBJECTS ##################################
#########################################################################################

#Before javelin
def run_from_toml1(filename):
    import pypetal.pipeline as pl

    output_dir, filenames, line_names = get_toml_inputs(filename)
    run_arr = get_toml_modules(filename)
    param_arr = get_toml_params(filename, run_arr)

    res = pl.run_pipeline(output_dir, filenames, line_names,
                          run_drw_rej=run_arr[0], drw_rej_params=param_arr[1],
                          run_detrend=run_arr[1], detrend_params=param_arr[2],
                          run_pyccf=run_arr[2], pyccf_params=param_arr[3],
                          run_pyzdcf=run_arr[3], pyzdcf_params=param_arr[4],
                          run_pyroa=run_arr[4], pyroa_params=param_arr[5],
                          **param_arr[0])

    return res


#After javelin
def run_from_toml2(filename):
    import pypetal.pipeline as pl

    output_dir, _, line_names = get_toml_inputs(filename)
    run_arr = get_toml_modules(filename)
    param_arr = get_toml_params(filename, run_arr)

    if run_arr[-1]:
        res = pl.run_weighting(output_dir, line_names,
                               run_pyccf=run_arr[2], pyccf_params=param_arr[3],
                               run_pyroa=run_arr[4], pyroa_params=param_arr[5],
                               run_javelin=run_arr[5], javelin_params=param_arr[6],
                               weighting_params=param_arr[7],
                               **param_arr[0])

        return res

    else:
        return

#########################################################################################
################################ TO RUN MULTIPLE OBJECTS ################################
#########################################################################################

def run_all1(filenames):
    for f in filenames:
        _ = run_from_toml1(f)

    return

def run_all3(filenames):
    for f in filenames:
        _ = run_from_toml2(f)

    return



import glob
import multiprocessing as mp
import os
import pickle
import shutil
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from os import devnull

import numpy as np
import PyROA
from astropy.table import Table

##############################################################
####################### SILENCE OUTPUT #######################
##############################################################


@contextmanager
def suppress_stdout_stderr():
    """A context manager that redirects stdout and stderr to devnull"""
    with open(devnull, 'w') as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield (err, out)

##############################################################
######################### FIT CLASS ##########################
##############################################################

class MyFit:

    def __init__(self, output_dir):

        self.output_dir = output_dir

        with open(output_dir + 'samples.obj', 'rb') as f:
            self.samples = pickle.load(f)

        with open(output_dir + 'samples_flat.obj', 'rb') as f:
            self.samples_flat = pickle.load(f)

        with open(output_dir + 'Lightcurve_models.obj', 'rb') as f:
            self.models = pickle.load(f)

        with open(output_dir + 'X_t.obj', 'rb') as f:
            dat = pickle.load(f)

            self.t = dat[0]
            self.X = dat[1]
            self.X_errs = dat[2]


##############################################################
######################## ASSIST FUNCTIONS ####################
##############################################################

def get_samples_chunks(samples, nburn=0, add_var=False, delay_dist=False):

    """Split the samples up into individual lines. The burn-in samples will be removed and the walkers will be flattened.


    Parameters
    ----------

    samples : list of float
        The samples object output from PyROA. Can be found in the "samples.obj" file.

    nburn : int, optional
        The number of burn-in samples to discard. Default is 0.



    Returns
    -------

    samples_chunks : list of float
        The chunked sample data. If the input ``samples`` array has a shape (nchain, nwalker, nobj*nparam + extra), the output ``samples_chunked`` array will
        have a shape ( nobj+1, (nchain-nburn)*nwalker, nparam ).

    """

    chunk_size = 3
    if add_var:
        chunk_size += 1

    #Remove burn-in
    samples = samples[nburn:,:,:].copy()

    #Insert tau=0 for the continuum
    samples = np.insert( samples, 2, np.zeros( (samples.shape[0], samples.shape[1]) ), axis=2 )

    #Insert delta_i=0 for the continuum
    if delay_dist:
        chunk_size += 1
        samples = np.insert(samples, 3, np.zeros( (samples.shape[0], samples.shape[1]) ), axis=2 )

    #Flatten
    samples = samples.reshape( (-1, samples.shape[2]) ).T

    #Chunk
    nlc = (samples.shape[0] - 1)//chunk_size
    samples_chunks = []
    for i in range(nlc+1):
        samples_chunks.append( samples[i*chunk_size:(i+1)*chunk_size] )

    return samples_chunks



def save_lines(line_fnames, line_names, output_dir, objname=None, delimiter=None, subtract_mean=False, div_mean=False):

    """Save light curves as PyROA-readable files. Light curves will be saved
    as ASCII files with the format: x, y, yerr. The saved files will be named "{objname}_{line_name}.dat".



    Parameters
    ----------

    line_fnames : list of str
        The filenames of the light curves to be saved.

    line_names : list of str
        The names of the light curves to be saved.

    output_dir : str
        The directory to save the light curves to.

    objname : str, optional
        The name of the object. This will Default is 'pyroa'.

    delimiter : str, optional
        The delimiter to use when reading the light curves. Default is None.

    subtract_mean : bool, optional
        If True, the mean of the light curve will be subtracted. Default is False.

    div_mean : bool, optional
        If True, the light curve will be divided by the mean. This occurs before subtracting the mean.
        Default is False.




    Returns
    -------

    new_fnames : list of str
        The filenames of the saved light curves.

    """

    if objname is None:
        objname = 'pyroa'

    new_fnames = []

    for i in range(len(line_fnames)):
        x, y, yerr = np.loadtxt( line_fnames[i], unpack=True, usecols=[0,1,2], delimiter=delimiter )

        if div_mean:
            yerr /= np.mean(y)
            y /= np.mean(y)

        if subtract_mean:
            y -= np.mean(y)

        new_fnames.append( output_dir + objname + '_' + line_names[i] + '.dat' )
        np.savetxt( new_fnames[-1], np.array([x,y,yerr]).T )

    return new_fnames




def get_priors(fnames, laglim_in, subtract_mean=False, div_mean=False, together=False, delimiter=','):


    """Get the priors used for PyROA.


    Parameters
    ----------

    fnames : list of str
        The filenames of the light curves.

    laglim_in : list of float
        The bounds to search between for each light curve. This should be a list of lists,
        each list having an upper and lower lag bound.

    subtract_mean : bool, optional
        If True, the mean of the light curve will be subtracted. Default is False.

    div_mean : bool, optional
        If True, the light curve will be divided by the mean. This occurs before subtracting the mean.
        Default is False.

    together : bool, optional
        Whether or not to fit all light curves to the continuum in one fit. Default is False.




    Returns
    -------

    prior_arr : list of float
        The array of priors used for PyROA.

    """


    if together:

        lower_lim = []
        upper_lim = []
        for i in range(len(fnames)-1):
            lower_lim.append( laglim_in[i][0] )
            upper_lim.append( laglim_in[i][1] )

        laglim = [ np.min(lower_lim), np.max(upper_lim) ]



        std_vals = []
        for i in range(len(fnames)):
            _, y, yerr = np.loadtxt( fnames[i], unpack=True, usecols=[0,1,2], delimiter=delimiter )

            if div_mean:
                yerr /= np.mean(y)
                y /= np.mean(y)

            if subtract_mean:
                y -= np.mean(y)

            std_vals.append( np.std(y) )


        if div_mean:
            err_prior = [0., 10.]
        else:
            err_prior = [0., 10*np.max(std_vals)]

        a_prior = [.5, 2.]
        b_prior = [.5, 2.]
        tau_prior = laglim
        delta_prior = [5., 50.]


        return [a_prior, b_prior, tau_prior, delta_prior, err_prior]


    else:
        prior_arr = np.zeros(( len(fnames)-1, 5, 2 ))

        _, y_cont, yerr_cont = np.loadtxt( fnames[0], unpack=True, usecols=[0,1,2], delimiter=delimiter )
        if div_mean:
            yerr_cont /= np.mean(y_cont)
            y_cont /= np.mean(y_cont)

        if subtract_mean:
            y_cont -= np.mean(y_cont)


        for i in range(len(fnames)-1):
            _, y, yerr = np.loadtxt( fnames[i+1], unpack=True, usecols=[0,1,2], delimiter=',' )

            if div_mean:
                yerr /= np.mean(y)
                y /= np.mean(y)

                if subtract_mean:
                    y -= np.mean(y)

                #err
                prior_arr[i,4,0] = 0.
                prior_arr[i,4,1] = 10.

            else:

                if subtract_mean:
                    y -= np.mean(y)

                #err - extra error
                prior_arr[i,4,0] = 0.
                prior_arr[i,4,1] = 10.*np.max([ np.std(y), np.std(y_cont) ])



            #a - RMS of the LC as a frac of MAD of the LC
            prior_arr[i,0,0] = .5
            prior_arr[i,0,1] = 2.

            #b - Mean of the LC as a frac of the mean of the LC
            prior_arr[i,1,0] = .5
            prior_arr[i,1,1] = 2.

            #tau
            prior_arr[i,2,0] = laglim_in[i][0]
            prior_arr[i,2,1] = laglim_in[i][1]

            #delta - window function width
            prior_arr[i,3,0] = 5.
            prior_arr[i,3,1] = 50.

        return prior_arr


def move_output_files(file_dir, output_dir):
    os.makedirs(output_dir, exist_ok=True)

    file_dir = os.path.abspath(file_dir) + r'/'
    output_dir = os.path.abspath(output_dir) + r'/'

    files = glob.glob(file_dir + '*.obj')
    for f in files:
        shutil.move(f, output_dir)

    return


def samples2table(samples, line_names, nburn=0, add_var=False, delay_dist=False):

    """Convert the PyROA "samples.obj" file into an astropy.table.Table object.


    Parameters
    ----------

    samples : list of float
        The samples from the PyROA run.

    line_names : list of str
        The names of the light curves.

    nburn : int, optional
        The number of burn-in samples to discard. Default is 0.

    add_var : bool, optional
        Whether the PyROA run included an extra error term. Default is False.

    delay_dist : bool, optional
        Whether the PyROA run included a delay distribution. Default is False.



    Returns
    -------

    table_dict : dict of astropy.table.Table
        A dictionary of astropy.table.Table objects, one for each light curve. There will also be
        a Table for the windowing function width, under the key 'delta'.

    """


    #Works for together=True
    samples_chunks = get_samples_chunks(samples, nburn)
    table_dict = {}

    for i, name in enumerate(line_names[1:]):
        samp_i = samples_chunks[i]
        tab_i = Table()

        tab_i.add_column( 'A', samp_i[:,0] )
        tab_i.add_column( 'B', samp_i[:,1] )
        tab_i.add_column( 'tau', samp_i[:,2] )

        if add_var:
            tab_i.add_column( 'sigma', samp_i[:,3] )

        if delay_dist:
            tab_i.add_column( 'delta', samples_chunks[-1][:,4] )

        table_dict[name] = tab_i


    delta_tab = Table()
    delta_tab.add_column( 'delta', samples_chunks[-1][:,0] )
    table_dict['delta'] = delta_tab

    return table_dict


def get_all_samples(samples, nburn=0):

    #Remove burn-in
    samples = samples[nburn:,:,:].copy()

    #Flatten
    samples = samples.reshape( (-1, samples.shape[2]) ).T

    return samples


##############################################################
##############################################################
##############################################################

def run_pyroa(fnames, lc_dir, line_dir, line_names,
              nburn=10000, nchain=15000, lag_bounds=None,
              init_tau=None, init_delta=10, sig_level=100,
              together=True, subtract_mean=True, div_mean=False,
              add_var=False, delay_dist=False, psi_types='Gaussian',
              objname=None, prior_func=None, verbose=True):


    """Run PyROA for a number of input light curves.
    NOTE: This will assume a log-Gaussian for the delay distribution.


    Parameters
    ----------

    fnames : list of str
        A list of paths to the light curves to be used in PyROA. The first light curve will be assumed to be the continuum. Must be in CSV format.

    lc_dir : str
        The output directory to put the light curves for PyROA to use.

    line_dir : str, list of str
        The output directory to store the data products output from PyROA.
        If together=False, this should be a list of paths for each line. If together=True, this should be a single path.

    line_names : list of str
        A list of the line names corresponding to the light curves.

    nburn : int, optional
        The number of burn-in samples to discard. Default is 10000.

    nchain : int, optional
        The number of samples to get per walker. Default is 15000.

    lag_bounds : list of str, optional
        The range of lags to consider for PyROA, one set of bounds for each line (excluding the continuum). If ``None``, the baseline of
        each light curve will be used as the lag bounds. Default is ``None``.

    init_tau : list of float
        A list of inital values for the time delay (tau) for each light curve. If ``None``, will set the inital value to 10.0 for each line.
        Default is ``None``.

    sig_level : float
        The number to use for sigma clipping in PyROA. Default is 100.

    together : bool, optional
        Whether to fit all light curves together or not. If ``together=False``, the ``line_dirs`` argument must be set. Default is ``True``.

    subtract_mean : bool, optional
        Whether to subtract the mean from all light curves before using PyROA or not. Will occur after ``div_mean`` if set to ``True``. Default is ``True``.

    div_mean : bool, optional
        Whether to divide each light curve by its mean before using PyROA. Will occur before ``subtract_mean`` if set to ``True``. Default is ``False``.

    add_var : bool or list of bool, optional
        Whether or not to add additional uncertainty in the data, same as the PyROA argument. If ``together=False``, multiple values may be given for each line.
        If only one value is given, it will be assumed for all lines. Default is ``True``.

    delay_dist : bool or list of bool, optional
        Same as the ``delay_dist`` argument for PyROA. If ``together=False``, multiple values may be given for each line.
        If only one value is given, it will be assumed for all lines. Default is ``True``.

    objname : str, optional
        The name of the object, will be used for plot and for the saved PyROA light curve data. If ``None``, will be set to "pyroa".

    prior_func : function, optional
        A function to use to get the priors for PyROA. Must have the same arguments as ``pypetal.pyroa.utils.get_priors`` except for the ``delimiter`` argument. 
        If ``None``, will use the default priors. Default is ``None``.

    line_names : list of str
        A list of directories to place the output PyROA data for each of the lines (excluding the continuum). Must be set if ``together=False``, and will only be used in such a case.
        Default is ``None``.



    Returns
    -------

    fit : PyROA.Fit or list of pyROA.Fit
        The PyROA.Fit object output from PyROA. If ``together=False``, this will be an array of the ``Fit`` objects.


    """


    if objname is None:
        objname = 'pyroa'

    if init_tau is None:
        init_tau = np.full( len(fnames)-1, 10. )

    if isinstance(psi_types, str):
        psi_types = [psi_types] * ( len(fnames)-1 )

    #If no lag bounds given, use the baseline of the light curves
    if lag_bounds is None:
        lag_bounds = []

        x_cont, _, _ = np.loadtxt( fnames[0], unpack=True, usecols=[0,1,2], delimiter=',' )
        for i in range(1, len(fnames)):
            x, _, _ = np.loadtxt( fnames[i], unpack=True, usecols=[0,1,2], delimiter=',' )

            max_x = np.max([ np.max(x), np.max(x_cont) ])
            min_x = np.min([ np.min(x), np.min(x_cont) ])
            bl = max_x - min_x

            lag_bounds.append([-bl, bl])


    if (not together) & (len(fnames)==2) & isinstance(line_dir, str):
        line_dir = [line_dir]


    if isinstance(line_dir, str):
        os.makedirs(line_dir, exist_ok=True)
    elif isinstance(line_dir, list):
        for ldir in line_dir:
            os.makedirs(ldir, exist_ok=True)
    else:
        raise ValueError('line_dir must be a string or list of strings')


    os.makedirs(lc_dir, exist_ok=True)

    if prior_func is None:
        prior_arr = get_priors(fnames, lag_bounds, subtract_mean=subtract_mean, div_mean=div_mean, together=together, delimiter=',')
    else:
        prior_arr = prior_func(fnames, lag_bounds, subtract_mean=subtract_mean, div_mean=div_mean, together=together)

    _ = save_lines(fnames, line_names, lc_dir, objname=objname, subtract_mean=subtract_mean, div_mean=div_mean, delimiter=',')

    cwd = os.getcwd()

    if not together:

        assert isinstance(line_dir, list), 'Must provide multiple line_dir if together=False'

        if isinstance(add_var, bool):
            add_var = np.full( len(fnames)-1, add_var )

        if isinstance(delay_dist, bool):
            delay_dist = np.full( len(fnames)-1, delay_dist )


        fit_arr = []

        for i in range(len(fnames)-1):

            filters = [line_names[0], line_names[i+1]]

            args = (lc_dir, objname, filters, prior_arr[i,:,:],)
            kwargs = {'add_var':add_var[i], 'init_tau':[init_tau[i]], 'init_delta':init_delta, 'sig_level':sig_level,
                      'delay_dist':delay_dist[i], 'psi_types':[psi_types[i]], 'Nsamples':nchain, 'Nburnin':nburn}

            if verbose:
                proc = mp.get_context('fork').Process(target=PyROA.Fit, args=args, kwargs=kwargs)

                proc.start()
                while proc.is_alive():
                    proc.is_alive()
                proc.terminate()

            else:
                with suppress_stdout_stderr():
                    proc = mp.get_context('fork').Process(target=PyROA.Fit, args=args, kwargs=kwargs)

                    proc.start()
                    while proc.is_alive():
                        proc.is_alive()
                    proc.terminate()


            move_output_files(cwd, line_dir[i])
            fit = MyFit(line_dir[i])

            fit_arr.append(fit)

        return fit_arr

    else:

        assert isinstance(line_dir, str), 'Must provide one line_dir if together=True'


        args = (lc_dir, objname, line_names, prior_arr,)
        kwargs = {'add_var':add_var, 'init_tau':init_tau, 'init_delta':init_delta, 'sig_level':sig_level,
                  'delay_dist':delay_dist, 'psi_types':psi_types, 'Nsamples':nchain, 'Nburnin':nburn}

        if verbose:
            proc = mp.get_context('fork').Process(target=PyROA.Fit, args=args, kwargs=kwargs)

            proc.start()
            while proc.is_alive():
                proc.is_alive()
            proc.terminate()

        else:
            with suppress_stdout_stderr():
                proc = mp.get_context('fork').Process(target=PyROA.Fit, args=args, kwargs=kwargs)

                proc.start()
                while proc.is_alive():
                    proc.is_alive()
                proc.terminate()

        move_output_files(cwd, line_dir)
        fit = MyFit(line_dir)

        return fit



import glob
import os
import warnings

import numpy as np

from pypetal.utils import defaults
from pypetal.utils.load import get_ordered_line_names
from pypetal.weighting.plotting import plot_weight_output, plot_weights
from pypetal.weighting.utils import (combine_weight_outputs,
                                     run_weighting_single)


def run_weighting_tot(output_dir,
                      jav_chain_fnames=None, pyccf_iccf_fnames=None, pyccf_dist_fnames=None,
                      pyroa_sample_fnames=None,
                      line_names=None, interp=2, together_jav=False,
                      pyroa_obj_inds=None, pyroa_params={},
                      general_kwargs={}, weighting_params={}, share_lag_bounds=True):


    """Run the weighting module for a number of lines, with pyPetal (and possible pyPetal-jav) already run.
    This will use the weighting scheme described in Grier et al. (2019) for all lag distributions provided.
    NOTE: If filenaes aren't given for a certain module (i.e. its input is None), it will be assumed not to have run.


    Parameters
    ----------

    output_dir : str
        The directory where the output will be saved. Must be the same output directory used for the pyPetal run.

    jav_chain_fnames : list of str, optional
        The filenames of the JAVELIN chains for each line. If None, JAVELIN will be assumed not to have run.
        Default is None.

    pyccf_iccf_fnames : list of str, optional
        The filenames of the PyCCF ICCF outputs for each line. If None, PyCCF will be assumed not to have run.
        Default is None.

    pyccf_dist_fnames : list of str, optional
        The filenames of the PyCCF lag distributions for each line. If None, PyCCF will be assumed not to have run.
        Default is None.

    pyroa_sample_fnames : list of str, optional
        The filenames of the PyROA sample for each line (i.e. "samples.obj"). If None, PyROA will be assumed not to have run.
        Default is None.

    line_names : list of str, optional
        The names of the lines to be weighted. If None, the output directory will be searched to determine the
        line_names. Default is None.

    interp : int, optional
        The interpolation factor to use for the PyCCF lag distributions. Default is 2.

    together_jav : bool, optional
        The together parameter used for the JAVELIN run. Default is False.

    pyroa_obj_inds : list of int, optional
        The indices in the samples_chunked array for each line (in the same order as line_names).
        Typically, the index for a given line will be one more than its index in line_names.
        Must be specified if pyroa_sample_fnames is not None. Default is None.

    pyroa_params : dict, optional
        The parameters used for the PyROA run. Default is {}.

    general_kwargs : dict, optional
        The general kwargs used for the pyPetal run. Default is {}.

    weighting_params : dict, optional
        The parameters to be used for the weighting. Default is {}.

    share_lag_bounds : bool, optional
        Whether to share the lag bounds for all lines when computing the weights. Default is True.


    Returns
    -------

    res_tot : list of dict
        A list of dictionaries containing the results for each line.

    summary_dicts : list of dict
        A list of dictionaries containing the summary information for each line. This information will
        be stored in the weight summary file in the output directory as well.

    """

    output_dir = os.path.abspath(output_dir) + r'/'
    pyccf_params = {'interp':interp}

    if line_names is None:
        warnings.warn('Assuming that the filenames are in the same order as the line names. Line names will be acquired in chronological order from the given directory, except the first will be the continuum', RuntimeWarning)
        line_names = get_ordered_line_names(output_dir)

    _, _, _, _, _, _, _, _, together_pyroa, _ = defaults.set_pyroa( pyroa_params, len(line_names) )

    #---------------------------
    #Get data fnames

    if output_dir + 'processed_lcs/' in glob.glob( output_dir + '*' ):
        line_fnames = np.array([ output_dir + 'processed_lcs/' + x + '_data.dat' for x in line_names ])
    else:
        line_fnames = np.array([ output_dir + 'light_curves/' + x + '.dat' for x in line_names ])


    general_kwargs = defaults.set_general(general_kwargs, line_fnames)
    _, _, _, _, zoom = defaults.set_weighting(weighting_params)

    #---------------------------
    #Share lag bounds?
    if share_lag_bounds:

        baselines = []

        x_cont, _, _ = np.loadtxt(line_fnames[0], unpack=True, delimiter=',', usecols=[0,1,2])
        for i in range(len(line_fnames)):
            x_line, _, _ = np.loadtxt(line_fnames[i], unpack=True, delimiter=',', usecols=[0,1,2])

            bl = np.max([ x_cont.max(), x_line.max() ]) - np.min([ x_cont.min(), x_line.min() ])
            baselines.append(bl)



        lag_bounds = []
        lag_bounds_i = [-np.max(baselines), np.max(baselines)]
        for i in range(len(line_fnames)-1):
            lag_bounds.append(lag_bounds_i)


    else:
        lag_bounds = general_kwargs['lag_bounds']


    #---------------------------
    #Account for None inputs
    run_javelin = True
    run_pyccf = True
    run_pyroa = True

    if jav_chain_fnames is None:
        warnings.warn('Assuming JAVELIN was not run.', RuntimeWarning)

    if (pyccf_iccf_fnames is None) or (pyccf_iccf_fnames is None):
        warnings.warn('Assuming PyCCF was not run.', RuntimeWarning)


    if jav_chain_fnames is None:
        jav_chain_fnames = [None] * ( len(line_names)-1)
        run_javelin = False
    elif together_jav:
        jav_chain_fnames = [jav_chain_fnames] * ( len(line_names)-1)
        run_javelin = True


    if pyccf_iccf_fnames is None:
        pyccf_iccf_fnames = [None] * ( len(line_names)-1)
        run_pyccf = False

    if pyccf_dist_fnames is None:
        pyccf_dist_fnames = [None] * ( len(line_names)-1)
        run_pyccf = False



    #Look for PyROA files if none input
    if pyroa_sample_fnames is None:
        warnings.warn('Looking for PyROA files, no files input', RuntimeWarning)


        if together_pyroa:
            fnames_i = glob.glob(output_dir + 'pyroa/*.obj')
            if len(fnames_i) == 0:
                warnings.warn('No PyROA files found. Assuming PyROA was not run.', RuntimeWarning)
                run_pyroa = False
            else:
                pyroa_sample_fnames = output_dir + 'pyroa/samples.obj'


        else:

            pyroa_sample_fnames = []
            for i in range(len(line_fnames[1:])):
                fnames_i = glob.glob(output_dir + r'/' + line_names[i+1] + 'pyroa/*.obj')

                if len(fnames_i) == 0:
                    warnings.warn('No PyROA files found for ' + line_names[i+1], RuntimeWarning)
                    pyroa_sample_fnames.append( None )

                else:
                   pyroa_sample_fnames.append( output_dir + r'/' + line_names[i+1] + 'pyroa/samples.obj' )

            if np.all([ x is None for x in pyroa_sample_fnames ]):
                warnings.warn('No PyROA files found. Assuming PyROA was not run.', RuntimeWarning)
                run_pyroa = False
                pyroa_sample_fnames = None


    #Check if PyROA files exist
    if isinstance(pyroa_sample_fnames, list):
        for i in range(len(pyroa_sample_fnames)):

            if os.path.isfile(pyroa_sample_fnames[i]):
                continue
            else:
               warnings.warn( pyroa_sample_fnames[i] + ' not found.', RuntimeWarning )
               pyroa_sample_fnames[i] = None

        if np.all([ x is None for x in pyroa_sample_fnames ]):
            warnings.warn('No PyROA files found. Assuming PyROA was not run.', RuntimeWarning)
            run_pyroa = False
            pyroa_sample_fnames = None


    elif isinstance(pyroa_sample_fnames, str):
        if not os.path.isfile(pyroa_sample_fnames):
            warnings.warn( pyroa_sample_fnames + ' not found. Assuming PyROA was not run.', RuntimeWarning )
            run_pyroa = False
            pyroa_sample_fnames = None


    if pyroa_sample_fnames is None:
        pyroa_sample_fnames = [None] * ( len(line_names)-1)
        run_pyroa=False
    elif together_pyroa:
        pyroa_sample_fnames = [pyroa_sample_fnames] * ( len(line_names)-1)
        run_pyroa = True


    if run_pyroa:
        if pyroa_obj_inds is None:
            if together_pyroa:
                pyroa_obj_inds = range(1, len(line_names))
            else:
                pyroa_obj_inds = []

                for i in range(len(line_names)-1):
                    if pyroa_sample_fnames[i] is None:
                        pyroa_obj_inds.append(None)
                    else:
                        pyroa_obj_inds.append(1)
    else:
        pyroa_obj_inds = [None] * ( len(line_names)-1)

    #---------------------------
    #Make weights directories
    for name in line_names[1:]:
        os.makedirs(output_dir + name + r'/weights/', exist_ok=True)

    #---------------------------
    #Run weighting

    summary_dicts = []
    outputs = []

    summary_fnames = []


    for i in range(len(line_fnames)-1):

        if together_jav:
            javelin_lag_col = 2 + 3*i
        else:
            javelin_lag_col = 2

        res, summary_dict = run_weighting_single(output_dir + line_names[i+1] + r'/weights/',
                                                 line_fnames[0], line_fnames[i+1],
                                                 weighting_params, lag_bounds[i],
                                                 jav_chain_fnames[i], pyccf_iccf_fnames[i], pyccf_dist_fnames[i],
                                                 pyroa_sample_fnames[i],
                                                 javelin_lag_col=javelin_lag_col,
                                                 pyroa_obj_ind=pyroa_obj_inds[i],
                                                 pyccf_params=pyccf_params, pyroa_params=pyroa_params)

        summary_dicts.append(summary_dict)
        outputs.append(res)
        summary_fnames.append(output_dir + name + r'/weights/weight_summary.fits')



    #---------------------------
    #Get total results

    res_tot = combine_weight_outputs(outputs, run_pyccf, run_javelin, run_pyroa)

    plot_mod = 'pyccf'
    if not run_pyccf:
        if run_javelin:
            plot_mod = 'javelin'
        elif run_pyroa:
            plot_mod = 'pyroa'

    for i in range(len(line_fnames)-1):
        plot_weights(output_dir, line_names[i+1], res_tot[plot_mod],
                        summary_dicts[i]['n0_'+plot_mod], summary_dicts[i]['k'],
                        general_kwargs['time_unit'], general_kwargs['plot'])

    if run_pyccf:
        plot_weight_output(output_dir, line_fnames[0], line_fnames, line_names,
                            res_tot['pyccf'], general_kwargs, 'pyccf', zoom,
                            general_kwargs['plot'])

    if run_javelin:
        plot_weight_output(output_dir, line_fnames[0], line_fnames, line_names,
                            res_tot['javelin'], general_kwargs, 'javelin', zoom,
                            general_kwargs['plot'])

    if run_pyroa:
        plot_weight_output(output_dir, line_fnames[0], line_fnames, line_names,
                            res_tot['pyroa'], general_kwargs, 'pyroa', zoom,
                            general_kwargs['plot'])


    return res_tot, summary_dicts



import itertools
import multiprocessing as mp
from functools import partial

import astropy.units as u
import numpy as np

from pypetal.drw_rej.utils import drw_flag
from pypetal.utils import defaults
from pypetal.utils.petalio import write_data


#For multiprocessing
def mp_map(func, args, threads):
    n_inputs = len(args)

    if (threads > 1) & (n_inputs > 1):
        pool = mp.get_context('fork').Pool(threads)
        res = pool.starmap(func, args)

        pool.close()
        pool.join()

    else:
        res = list(itertools.starmap(func, args))

    return res



def str2unit(x):

    if (x == 'Arbitrary Units'):
        unit = u.dimensionless_unscaled
    else:
        unit = u.Unit(x)

    return unit


def drw_rej_tot(cont_fname, line_fnames, line_names, output_dir,
                general_kwargs, kwargs):


    #--------------------------------------------------
    #Read general kwargs

    verbose = general_kwargs['verbose']
    plot = general_kwargs['plot']
    time_unit = general_kwargs['time_unit']
    lc_unit = general_kwargs['lc_unit']
    threads = general_kwargs['threads']

    #--------------------------------------------------
    #Read kwargs

    jitter, nsig, nwalker, nburn, nchain, clip, \
        reject_data, use_for_javelin = defaults.set_drw_rej(kwargs,
                                                            np.hstack([ [cont_fname], line_fnames ])
                                                            )

    #--------------------------------------------------
    #Get units

    time_unit = str2unit(time_unit)

    for i in range(len(lc_unit)):
        lc_unit[i] = str2unit(lc_unit[i])

    #--------------------------------------------------
    if verbose:

        if len(clip) > 2:
            clip_str = 'array'
        else:
            clip_str = clip

        txt_str = """
Performing DRW rejection
------------------------
jitter: {}
nsig: {}
nwalker: {}
nburn: {}
nchain: {}
clip: {}
reject_data: {}
use_for_javelin: {}
------------------------
        """.format( jitter, nsig, nwalker, nburn,
                    nchain, clip_str, reject_data, use_for_javelin)

        print(txt_str)


    sigmas = []
    taus = []

    if jitter:
        jitters = []

    rejecting = np.any(reject_data)
    fnames = np.hstack([ [cont_fname], line_fnames ])


    drw_rej_func = partial( drw_flag, nwalkers=nwalker, nburn=nburn, nsamp=nchain,
                            nsig=nsig, jitter=jitter, plot=plot)



    arg1 = []
    arg2 = []
    arg3 = []
    arg4 = line_names
    arg5 = []

    for i in range(len(fnames)):
        if reject_data[i]:
            x, y, yerr = np.loadtxt( fnames[i], delimiter=',', unpack=True, usecols=[0,1,2] )

            arg1.append( x*time_unit )
            arg2.append( y*lc_unit[i] )
            arg3.append( yerr*lc_unit[i] )
            arg5.append( output_dir + line_names[i] + '/drw_rej/' + line_names[i] + '_drw_fit.pdf' )


    args = list(zip(arg1, arg2, arg3, arg4, arg5))
    res_tot = mp_map(drw_rej_func, args, threads)

    masks_tot = []
    n = 0
    for i in range(len(fnames)):

        if reject_data[i]:
            res = res_tot[n]

            sigmas.append( res['sigma'])
            taus.append( res['tau'])

            if jitter:
                jitters.append( res['jitter'] )

            mask = res['mask']

            #Write mask
            dat_fname = output_dir + line_names[i] + '/drw_rej/' + line_names[i] + '_mask.dat'
            write_data(mask, dat_fname)

            #Write LC fit
            dat_fname = output_dir + line_names[i] + '/drw_rej/' + line_names[i] + '_drw_fit.dat'
            header = '#time,light curve,error'
            write_data( [ res['fit_x'], res['fit_y'], res['fit_err'] ], dat_fname, header )

            #Write the MCMC chains
            dat_fname = output_dir + line_names[i] + '/drw_rej/' + line_names[i] + '_chain.dat'

            if jitter:
                header = '#sigma,tau,jitter'
                write_data( [ res['sigma'], res['tau'], res['jitter'] ], dat_fname, header )
            else:
                header = '#sigma,tau'
                write_data( [ res['sigma'], res['tau'] ], dat_fname, header )


            x_new = x[~mask]
            y_new = y[~mask]
            yerr_new = yerr[~mask]

            dat_fname = output_dir + 'processed_lcs/' + line_names[i] + '_data.dat'
            write_data( [x_new, y_new, yerr_new], dat_fname )


            n += 1

        else:
            x, y, yerr = np.loadtxt( fnames[i], delimiter=',', unpack=True, usecols=[0,1,2] )

            mask = np.full( len(x), False )
            taus.append(None)
            sigmas.append(None)

            if jitter:
                jitters.append(None)

            if rejecting:
                dat_fname = output_dir + 'processed_lcs/' + line_names[i] + '_data.dat'
                write_data([x, y, yerr], dat_fname)


        masks_tot.append(mask)

    output = {
        'masks': masks_tot,
        'reject_data': reject_data,
        'taus': taus,
        'sigmas': sigmas
    }

    if jitter:
        output['jitters'] = jitters

    return output



import itertools
import multiprocessing as mp

import numpy as np

from pypetal.pyccf.plotting import plot_pyccf_results
from pypetal.pyccf.utils import get_pyccf_lags
from pypetal.utils import defaults
from pypetal.utils.petalio import write_data


#For multiprocessing
def mp_map(func, args, threads):
    n_inputs = len(args)

    if (threads > 1) & (n_inputs > 1):
        pool = mp.get_context('fork').Pool(threads)
        res = pool.starmap(func, args)

        pool.close()
        pool.join()

    else:
        res = list(itertools.starmap(func, args))

    return res



def pyccf_tot(cont_fname, line_fnames, line_names, output_dir,
              general_kwargs, kwargs):

    if line_fnames is str:
        line_fnames = [line_fnames]

    #--------------------------------------------------
    #Read general kwargs

    verbose = general_kwargs['verbose']
    plot = general_kwargs['plot']
    time_unit = general_kwargs['time_unit']
    lc_unit = general_kwargs['lc_unit']
    lag_bounds = general_kwargs['lag_bounds']
    threads = general_kwargs['threads']

    #--------------------------------------------------
    #Read kwargs
    interp, nsim, mcmode, sigmode, thres, nbin = defaults.set_pyccf(kwargs)

    #-------------------------------------------
    #Run pyCCF for each line

    if verbose:

        if len(lag_bounds) > 2:
            lag_bounds_str = 'array'
        else:
            lag_bounds_str = lag_bounds

        txt_str = """
Running pyCCF
-----------------
lag_bounds: {}
interp: {}
nsim: {}
mcmode: {}
sigmode: {}
thres: {}
nbin: {}
-----------------
        """.format( lag_bounds_str, interp, nsim,
                    mcmode, sigmode, thres, nbin)

        print(txt_str)


    res_tot = []
    for i in range(len(line_fnames)):
        res = get_pyccf_lags( cont_fname, line_fnames[i], lag_bounds[i],
                                   interp=interp, nsim=nsim,
                                   mcmode=mcmode, sigmode=sigmode, thres=thres,
                                   threads=threads, verbose=verbose)

        #Write CCF to file
        dat_fname = output_dir + line_names[i+1] + r'/pyccf/' + line_names[i+1] + '_ccf.dat'
        header = '#Lag,CCF'
        write_data( [ res['CCF_lags'], res['CCF'] ], dat_fname, header )

        #Write CCCD, CCPD to file
        dat_fname = output_dir + line_names[i+1] + r'/pyccf/' + line_names[i+1] + '_ccf_dists.dat'
        header = '#CCCD,CCPD'
        write_data( [ res['CCCD_lags'], res['CCPD_lags'] ], dat_fname, header )


        res['name'] = line_names[i+1]

        x1, y1, yerr1 = np.loadtxt( cont_fname, delimiter=',', unpack=True, usecols=[0,1,2] )
        x2, y2, yerr2 = np.loadtxt( line_fnames[i], delimiter=',', unpack=True, usecols=[0,1,2] )

        plot_pyccf_results(x1, y1, yerr1, x2, y2, yerr2,
                                    res['CCF_lags'], res['CCF'],
                                    res['CCCD_lags'], res['CCPD_lags'],
                                    nbin=nbin, time_unit=time_unit, lc_unit=[lc_unit[0], lc_unit[i+1]],
                                    lc_names=[line_names[0], line_names[i+1]],
                                    fname = output_dir + line_names[i+1] + r'/pyccf/' + line_names[i+1] + '_ccf.pdf',
                                    show=plot)


        res_tot.append(res)


    return res_tot
